using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.Text;
using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Graphics;

namespace Waboga
{
    public class PlayerPiece
    {
        public Point Position;
        public PlayerIndex Player;
        public int PieceIndex;

        public bool Finished
        {
            get
            {
                // Get the goal area
                Point goal_start, goal_dir;
                int goal_len;

                GetGoalPieces(Player, out goal_start, out goal_dir, out goal_len);

                // Get distance to the goal and test (assumes that the goal is on the primary axis and takes all boundary pieces)
                if (goal_dir.X == 0 && Position.X == goal_start.X)
                    return true;
                if (goal_dir.Y == 0 && Position.Y == goal_start.Y)
                    return true;

                return false;
            }
        }

        public PlayerPiece(PlayerIndex plr, int piece_idx, int total_pieces)
        {
            Player = plr;
            PieceIndex = piece_idx;

            // Get the start position
            Point pos_start, pos_dir;
            int pos_len;

            switch (plr)
            {
                default:
                case PlayerIndex.One:
                    pos_start = new Point(0, Board.Height - 1);
                    pos_dir = new Point(1, 0);
                    pos_len = Board.Width;
                    break;

                case PlayerIndex.Two:
                    pos_start = new Point(0, 0);
                    pos_dir = new Point(1, 0);
                    pos_len = Board.Width;
                    break;

                case PlayerIndex.Three:
                    pos_start = new Point(0, 0);
                    pos_dir = new Point(0, 1);
                    pos_len = Board.Height;
                    break;

                case PlayerIndex.Four:
                    pos_start = new Point(Board.Width - 1, 0);
                    pos_dir = new Point(0, 1);
                    pos_len = Board.Height;
                    break;
            }

            int pos = (int)((float)(1 + piece_idx) * ((float)pos_len / (float)(total_pieces + 1)));
            Position = PointUtil.Add(pos_start, PointUtil.Scale(pos_dir, pos));
        }

        public static void GetGoalPieces(PlayerIndex plr, out Point start_pos, out Point dir, out int len)
        {
            switch (plr)
            {
                default:
                case PlayerIndex.One:
                    start_pos = new Point(0, 0);
                    dir = new Point(1, 0);
                    len = Board.Width;
                    break;

                case PlayerIndex.Two:
                    start_pos = new Point(0, Board.Height - 1);
                    dir = new Point(1, 0);
                    len = Board.Width;
                    break;

                case PlayerIndex.Three:
                    start_pos = new Point(Board.Width - 1, 0);
                    dir = new Point(0, 1);
                    len = Board.Height;
                    break;

                case PlayerIndex.Four:
                    start_pos = new Point(0, 0);
                    dir = new Point(0, 1);
                    len = Board.Height;
                    break;
            }
        }
    }

    public class BoardState
    {
        private List<Point> m_pieces = new List<Point>();

        private bool[,] m_wall_horiz = new bool[Board.Width, Board.Height + 1];
        private bool[,] m_wall_vert = new bool[Board.Width + 1, Board.Height];
        private bool[,] m_filled_intersection = new bool[Board.Width + 1, Board.Height + 1];

        public void CopyFrom(Board b)
        {
            m_pieces.Clear();
            List<PlayerPiece> pieces = b.pieces;
            for (int i = 0; i < pieces.Count; ++i)
                m_pieces.Add(pieces[i].Position);

            Array.Copy(b.wall_horiz, m_wall_horiz, m_wall_horiz.Length);
            Array.Copy(b.wall_vert, m_wall_vert, m_wall_vert.Length);
            Array.Copy(b.filled_intersection, m_filled_intersection, m_filled_intersection.Length);
        }

        public void CopyFrom(BoardState b)
        {
            m_pieces.Clear();
            for (int i = 0; i < b.m_pieces.Count; ++i)
                m_pieces.Add(b.m_pieces[i]);

            Array.Copy(b.m_wall_horiz, m_wall_horiz, m_wall_horiz.Length);
            Array.Copy(b.m_wall_vert, m_wall_vert, m_wall_vert.Length);
            Array.Copy(b.m_filled_intersection, m_filled_intersection, m_filled_intersection.Length);
        }

        public void CopyTo(Board b)
        {
            Debug.Assert(m_pieces.Count == b.pieces.Count);
            for (int i = 0; i < m_pieces.Count; ++i)
                b.pieces[i].Position = m_pieces[i];

            Array.Copy(m_wall_horiz, b.wall_horiz, m_wall_horiz.Length);
            Array.Copy(m_wall_vert, b.wall_vert, m_wall_vert.Length);
            Array.Copy(m_filled_intersection, b.filled_intersection, m_filled_intersection.Length);
        }
    }

    public class Board
    {
        public const int Width = 11;
        public const int Height = 11;

        public int PlayerCount
        {
            get { return m_player_count; }
        }

        public int PieceCount
        {
            get { return m_piece_count; }
        }

        public bool CanBeCompleted
        {
            get
            {
                for (int plr = 0; plr < PlayerCount; ++plr)
                {
                    if (!CanPlayerComplete((PlayerIndex)plr))
                        return false;
                }
                return true;
            }
        }

        public bool HasGameCompleted
        {
            get
            {
                PlayerIndex? unfinished = null;
                for (int i = 0; i < pieces.Count; ++i)
                {
                    PlayerPiece piece = pieces[i];
                    if (!piece.Finished)
                    {
                        if (unfinished != null && unfinished != piece.Player)
                            return false;
                        unfinished = piece.Player;
                    }
                }
                return true;
            }
        }

        private int m_player_count;
        private int m_piece_count;

        public bool[,] wall_horiz = new bool[Width, Height + 1];
        public bool[,] wall_vert = new bool[Width + 1, Height];
        public bool[,] filled_intersection = new bool[Width + 1, Height + 1];

        private BoardState m_undo = new BoardState();

        public List<PlayerPiece> pieces;

        public WallPiece active_edit_piece;

        public PlayerPiece active_player_piece;
        public List<Point> active_player_piece_destinations = new List<Point>();
        public Point active_player_piece_position;

        public Board()
        {
            pieces = new List<PlayerPiece>();
        }

        public void Init(int plrs, int piece_cnt)
        {
            m_player_count = plrs;
            m_piece_count = piece_cnt;

            pieces.Clear();
            for (int i = 0; i < plrs; ++i)
            {
                for (int piece = 0; piece < piece_cnt; ++piece)
                {
                    pieces.Add(new PlayerPiece((PlayerIndex)i, piece, piece_cnt));
                }
            }

            Array.Clear(wall_horiz, 0, wall_horiz.Length);
            Array.Clear(wall_vert, 0, wall_vert.Length);
            Array.Clear(filled_intersection, 0, filled_intersection.Length);

            active_edit_piece = null;
            active_player_piece = null;
            active_player_piece_destinations.Clear();

            SaveUndo();
        }

        public void DoUndo()
        {
            m_undo.CopyTo(this);
        }

        public void SaveUndo()
        {
            m_undo.CopyFrom(this);
        }

        public bool HasPlayerCompleted(PlayerIndex plr)
        {
            for (int i = 0; i < pieces.Count; ++i)
            {
                PlayerPiece piece = pieces[i];
                if (piece.Player == plr && !piece.Finished)
                    return false;
            }
            return true;
        }

        public PlayerPiece GetPiece(PlayerIndex plr, int index)
        {
            for (int i = 0; i < pieces.Count; ++i)
            {
                PlayerPiece piece = pieces[i];
                if (piece.Player == plr && piece.PieceIndex == index)
                    return piece;
            }
            return null;
        }

        public bool DoesHitWall(Point from, Direction dir)
        {
            // Test that the place we're moving to is inside the field
            Point new_pos = PointUtil.Add(from, DirectionUtil.ToPoint(dir));
            if (new_pos.X < 0 || new_pos.Y < 0 || new_pos.X >= Width || new_pos.Y >= Height)
                return true;

            // Test a wall
            bool[,] test_array;
            Point test_pos = from;

            switch (dir)
            {
                default:
                case Direction.Left:
                    test_array = wall_vert;
                    break;

                case Direction.Right:
                    test_array = wall_vert;
                    test_pos.X += 1;
                    break;

                case Direction.Up:
                    test_array = wall_horiz;
                    break;

                case Direction.Down:
                    test_array = wall_horiz;
                    test_pos.Y += 1;
                    break;
            }

            return test_array[test_pos.X, test_pos.Y];
        }

        public PlayerIndex? GetPlayerAt(Point pos)
        {
            for (int i = 0; i < pieces.Count; ++i)
            {
                if (pieces[i].Position == pos)
                    return pieces[i].Player;
            }
            return null;
        }

        public void GetPieceDestinations(PlayerPiece piece, List<Point> dest_points)
        {
            Point piece_pos = piece.Position;
            PlayerIndex piece_plr = piece.Player;

            dest_points.Clear();

            // Test each direction
            for (int diri = 0; diri < 4; ++diri)
            {
                TestMoveDir(piece_plr, piece_pos, (Direction)diri, dest_points);
            }
        }

        private void TestMoveDir(PlayerIndex plr, Point from, Direction dir, List<Point> dest_points)
        {
            // Hits wall?
            if (DoesHitWall(from, dir))
                return;

            // Move to the tile
            PointUtil.Add(ref from, DirectionUtil.ToPoint(dir));

            // A player here?
            PlayerIndex? who_is_here = GetPlayerAt(from);
            if (who_is_here != null)
            {
                // If it's an own piece, ignore it
                if (who_is_here == plr)
                    return;

                // Enemy piece: we can either jump over it, or else go to sides
                if (DoesHitWall(from, dir))
                { // Must test sides
                    Direction side1 = DirectionUtil.GetSide(dir, false);
                    Direction side2 = DirectionUtil.GetSide(dir, true);

                    TestMoveDir(plr, from, side1, dest_points);
                    TestMoveDir(plr, from, side2, dest_points);
                }
                else
                { // Can jump over
                    TestMoveDir(plr, from, dir, dest_points);
                }
                return;
            }

            // Place is valid, add if not found in the list already
            if (!dest_points.Contains(from))
                dest_points.Add(from);
        }

        private Stack<Point> m_testfill_todo = new Stack<Point>();
        private int[,] m_testfill_table = new int[Width, Height];

        public int DistanceToHorizLine(int dest_y, PlayerPiece piece)
        {
            // Init the starting tables
            m_testfill_todo.Clear();
            Array.Clear(m_testfill_table, 0, m_testfill_table.Length);

            // Set the start conditions
            Point start_pos = piece.Position;
            m_testfill_todo.Push(start_pos);
            m_testfill_table[start_pos.X, start_pos.Y] = 1;

            // Floodfill until no more entries are left
            while (m_testfill_todo.Count > 0)
            {
                Point cur_pos = m_testfill_todo.Pop();
                int val_cur = m_testfill_table[cur_pos.X, cur_pos.Y];

                // Already at it?
                /*if (cur_pos.Y == dest_y)
                    continue;
                // Route too long?
                if (val_cur > (Width * 1))
                    continue;*/

                // Test every direction
                for (int diri = 0; diri < 4; ++diri)
                {
                    Direction dir = (Direction)diri;

                    if (DoesHitWall(cur_pos, dir))
                        continue;

                    // Found a valid direction
                    Point new_pos = PointUtil.Add(cur_pos, DirectionUtil.ToPoint(dir));

                    int val_at = m_testfill_table[new_pos.X, new_pos.Y];
                    if (val_at == 0 || (val_cur + 1) < val_at)
                    {
                        m_testfill_table[new_pos.X, new_pos.Y] = (byte)(val_cur + 1);
                        m_testfill_todo.Push(new_pos);
                    }
                }
            }

            // Get the distance
            Point goal_pos;
            Point goal_dir;
            int goal_len;

            PlayerPiece.GetGoalPieces(piece.Player, out goal_pos, out goal_dir, out goal_len);

            int dist = Width * Height;
            /*for (int x = 0; x < Width; ++x)
            {
                byte v = m_testfill_table[x, dest_y];
                if (v > 0 && v < dist)
                    dist = v;
            }*/
            while (goal_len-- > 0)
            {
                int v = m_testfill_table[goal_pos.X, goal_pos.Y];
                if (v > 0 && v < dist)
                    dist = v;
                goal_pos = PointUtil.Add(goal_pos, goal_dir);
            }
            return dist - 1;
        }

        private bool CanPlayerComplete(PlayerIndex plr)
        {
            // Get the starting area (we start from the players finishing area and work backwards)
            Point start_pos;
            Point start_dir;
            int start_len;

            PlayerPiece.GetGoalPieces(plr, out start_pos, out start_dir, out start_len);

            // Init the starting tables
            m_testfill_todo.Clear();
            Array.Clear(m_testfill_table, 0, m_testfill_table.Length);

            for (int i = 0; i < start_len; ++i)
            {
                m_testfill_todo.Push(start_pos);
                m_testfill_table[start_pos.X, start_pos.Y] = 1;
                PointUtil.Add(ref start_pos, start_dir);
            }

            // Floodfill until no more entries are left
            while (m_testfill_todo.Count > 0)
            {
                Point cur_pos = m_testfill_todo.Pop();

                // Test every direction
                for (int diri = 0; diri < 4; ++diri)
                {
                    Direction dir = (Direction)diri;

                    if (DoesHitWall(cur_pos, dir))
                        continue;

                    // Found a valid direction
                    Point new_pos = PointUtil.Add(cur_pos, DirectionUtil.ToPoint(dir));

                    if (m_testfill_table[new_pos.X, new_pos.Y] == 0)
                    {
                        m_testfill_table[new_pos.X, new_pos.Y] = 1;
                        m_testfill_todo.Push(new_pos);
                    }
                }
            }

            // Check that every active player piece is on a reached node
            for (int i = 0; i < pieces.Count; ++i)
            {
                PlayerPiece piece = pieces[i];
                if (piece.Player != plr)
                    continue;

                if (m_testfill_table[piece.Position.X, piece.Position.Y] == 0)
                    return false;
            }

            return true;
        }
    }
}
