using System;
using System.Collections.Generic;
using System.Text;
using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Graphics;
using Microsoft.Xna.Framework.Input;

namespace Waboga
{
    public class BoardVisualizer
    {
        private const float PieceSize = 1.0f;
        private const float PieceScale = 1.0f / 100.0f;
        private const float CenterOffsetX = -(Board.Width * PieceSize) * 0.5f;
        private const float CenterOffsetY = -(Board.Height * PieceSize) * 0.5f;

        private const float CameraRotSpeed = 40.0f;
        private const float CameraTiltSpeed = 40.0f;
        private const float CameraRotSpeedMouse = 10.0f;
        private const float CameraTiltSpeedMouse = 10.0f;

        private const float PieceFloorY = 0.0f;
        private const float WallFloorY = 0.0f;

        private const float HiddenY = 16.0f;

        private static Matrix[] m_goal_marker_matrix = new Matrix[]
            {
                Matrix.CreateRotationX(MathHelper.ToRadians(-90.0f)),
                Matrix.CreateRotationX(MathHelper.ToRadians(-90.0f)) * Matrix.CreateRotationY(MathHelper.ToRadians(180.0f)),
                Matrix.CreateRotationX(MathHelper.ToRadians(-90.0f)) * Matrix.CreateRotationY(MathHelper.ToRadians(270.0f)),
                Matrix.CreateRotationX(MathHelper.ToRadians(-90.0f)) * Matrix.CreateRotationY(MathHelper.ToRadians(90.0f)),
            };

        private Board m_board;
        private Resources m_res;

        private float m_cam_dist = 15.0f;
        private float m_cam_ang_rot = 0.0f;
        private float m_cam_ang_tilt = 45.0f;

        private WallPiece m_showing_wall;
        private PlayerPiece m_showing_plr;
        private Vector3 m_showing_piece_pos;
        private float m_showing_piece_ang;

        private float m_color_wave_player;

        private float m_render_alpha;
        private Color? m_render_color;

        public Matrix CameraMatrix
        {
            get
            {
                return
                    Matrix.CreateRotationY(MathHelper.ToRadians(m_cam_ang_rot)) *
                    Matrix.CreateRotationX(MathHelper.ToRadians(m_cam_ang_tilt)) *
                    Matrix.CreateTranslation(new Vector3(0.0f, 0.0f, -m_cam_dist));
            }
        }

        public Matrix ProjisionMatrix
        {
            get
            {
                return
                    Matrix.CreatePerspectiveFieldOfView(
                        MathHelper.ToRadians(50),
                        (float)m_res.graphics.GraphicsDevice.DisplayMode.Width / (float)m_res.graphics.GraphicsDevice.DisplayMode.Height,
                        0.1f,
                        100.0f);
            }
        }

        public bool BoardVisible = false;

        public PlayerIndex? ActivePlayer;

        private struct VelObj
        {
            public float y;
            public float vel;

            public VelObj(float _y)
            {
                y = _y;
                vel = 0.0f;
            }

            public VelObj(float _y, float _vel)
            {
                y = _y;
                vel = _vel;
            }
        }

        private VelObj[,] m_vis_tile = new VelObj[Board.Width, Board.Height];
        private VelObj[,] m_vis_wall_horiz = new VelObj[Board.Width, Board.Height + 1];
        private VelObj[,] m_vis_wall_vert = new VelObj[Board.Width + 1, Board.Height];
        private VelObj[,] m_vis_isect = new VelObj[Board.Width + 1, Board.Height + 1];

        public BoardVisualizer(Board board, Resources res)
        {
            m_board = board;
            m_res = res;

            for (int y = 0; y <= Board.Height; ++y)
            {
                bool yok = y < Board.Height;
                for (int x = 0; x <= Board.Width; ++x)
                {
                    bool xok = x < Board.Width;

                    if (xok && yok)
                        m_vis_tile[x, y] = new VelObj(HiddenY);
                    if (xok)
                        m_vis_wall_horiz[x, y] = new VelObj(HiddenY);
                    if (yok)
                        m_vis_wall_vert[x, y] = new VelObj(HiddenY);
                    m_vis_isect[x, y] = new VelObj(HiddenY);
                }
            }
        }

        public Vector2 GetBoardFloatDirectionFromController(Vector2 ctrl_dir)
        {
            // Apply the dead zone
            if (ctrl_dir.Length() < 0.2f)
                return Vector2.Zero;

            // Get the direction in world coordinates
            Matrix cam_mtx = Matrix.Invert(CameraMatrix);
            Vector3 world_dir = cam_mtx.Right * ctrl_dir.X + cam_mtx.Up * ctrl_dir.Y;

            // Dismiss the Y direction
            world_dir.Y = 0.0f;
            if (world_dir.LengthSquared() < 0.01f)
                return Vector2.Zero;

            world_dir.Normalize();
            return new Vector2(world_dir.X, world_dir.Z);
        }

        public Point GetBoardDirectionFromController(Vector2 ctrl_dir)
        {
            Vector2 dir = GetBoardFloatDirectionFromController(ctrl_dir);
            Point ret = Point.Zero;

            if (Math.Abs(dir.X) >= 0.2f)
                ret.X = (dir.X >= 0.0f) ? 1 : -1;
            if (Math.Abs(dir.Y) >= 0.2f)
                ret.Y = (dir.Y >= 0.0f) ? 1 : -1;

            return ret;
        }

        public void Update(float delta_t)
        {
            if (BoardVisible)
            {
                Vector2 cam_stick = m_res.controller.RightStick;
                Vector2 mouse_drag = m_res.controller.MouseDragAmount;

                // Dead zone
                if (Math.Abs(cam_stick.X) < 0.2f)
                    cam_stick.X = 0.0f;
                if (Math.Abs(cam_stick.Y) < 0.2f)
                    cam_stick.Y = 0.0f;

                // Move the camera
                m_cam_ang_rot += cam_stick.X * delta_t * CameraRotSpeed;
                m_cam_ang_tilt += cam_stick.Y * delta_t * CameraTiltSpeed;

                m_cam_ang_rot += mouse_drag.X * delta_t * CameraRotSpeedMouse;
                m_cam_ang_tilt += mouse_drag.Y * delta_t * CameraTiltSpeedMouse;

                while (m_cam_ang_rot < 0.0f)
                    m_cam_ang_rot += 360.0f;
                while (m_cam_ang_rot >= 360.0f)
                    m_cam_ang_rot -= 360.0f;

                m_cam_ang_tilt = MathHelper.Clamp(m_cam_ang_tilt, 10.0f, 90.0f);
            }

            // Init the position update
            float dest_angle = m_showing_piece_ang;
            Vector3 dest_position = m_showing_piece_pos;

            // Update the showing wall piece
            if (m_board.active_edit_piece == null)
            {
                m_showing_wall = null;
            }
            else
            {
                bool new_piece = (m_showing_wall == null);
                m_showing_wall = m_board.active_edit_piece;

                // Get the current piece stats
                dest_angle = -((float)m_board.active_edit_piece.Angle * 360.0f / 4.0f);
                PiecePosition(out dest_position, m_showing_wall.Position.X, m_showing_wall.Position.Y);

                // Init a new piece positions? (no any previous wall piece)
                if (new_piece)
                {
                    m_showing_piece_pos = dest_position;
                    m_showing_piece_ang = dest_angle;
                }
            }

            // Update the showing player piece
            if (m_board.active_player_piece == null)
            {
                m_showing_plr = null;
            }
            else
            {
                bool new_piece = (m_showing_plr == null);
                m_showing_plr = m_board.active_player_piece;

                // Get the position                
                m_showing_piece_ang = dest_angle = 0.0f;
                PiecePosition(out dest_position, m_board.active_player_piece_position.X, m_board.active_player_piece_position.Y);

                // Init a new piece positions? (no any previous wall piece)
                if (new_piece)
                {
                    m_showing_piece_pos = dest_position;
                }
            }

            // Lerp the visible position/angles
            m_showing_piece_pos = Vector3.Lerp(m_showing_piece_pos, dest_position, 0.15f);

            m_showing_piece_ang += MathUtil.DegreeDistance(m_showing_piece_ang, dest_angle) * 0.15f;
            m_showing_piece_ang = MathUtil.WrapRange(m_showing_piece_ang, 0.0f, 360.0f);

            // Color wave
            m_color_wave_player = MathUtil.WrapRange(m_color_wave_player + delta_t * 2.0f, 0.0f, 2.0f * (float)Math.PI);

            // Update the visible pieces
            UpdateVisible(delta_t);
        }

        private void UpdateVisible(float delta_t)
        {
            for (int y = 0; y <= Board.Height; ++y)
            {
                bool yok = y < Board.Height;
                for (int x = 0; x <= Board.Width; ++x)
                {
                    bool xok = x < Board.Width;

                    if (xok && yok)
                        DoVisFall(x, y, BoardVisible, delta_t, ref m_vis_tile[x, y], false);
                    if (xok)
                    {
                        bool vis = BoardVisible && m_board.wall_horiz[x, y];
                        DoVisFall(x, y, vis, delta_t, ref m_vis_wall_horiz[x, y], true);
                    }
                    if (yok)
                    {
                        bool vis = BoardVisible && m_board.wall_vert[x, y];
                        DoVisFall(x, y, vis, delta_t, ref m_vis_wall_vert[x, y], true);
                    }
                    {
                        bool vis = BoardVisible && m_board.filled_intersection[x, y];
                        DoVisFall(x, y, vis, delta_t, ref m_vis_isect[x, y], true);
                    }
                }
            }
        }

        private bool DoVisFall(int x, int y, bool going_to_visible, float delta_t, ref VelObj dst, bool is_piece)
        {
            // Get the variation multiplier
            uint vari = (uint)(x + y * (Board.Width + 1)) + 1237657;
            vari = (uint)((ulong)vari * (ulong)1664525 + (ulong)1013904223);
            float acc_mul = 1.0f + (float)(vari % 128) * 0.01f;

            // Get the fall acceleration
            float acc = (is_piece ? 2.0f : 0.1f) * acc_mul;
            if (going_to_visible)
                acc = -acc;

            // Advance
            dst.vel += acc * delta_t;
            dst.y += dst.vel;

            // There yet?
            if (going_to_visible && dst.y <= 0.0f)
            {
                dst.y = 0.0f;
                dst.vel = 0.0f;
                return true;
            }
            if (!going_to_visible && dst.y >= HiddenY)
            {
                dst.y = HiddenY;
                dst.vel = 0.0f;
                return true;
            }
            return false;
        }

        public void Render(GraphicsDeviceManager graphics)
        {
            Matrix camera_mtx = CameraMatrix;
            //Matrix camera_mtx = Matrix.CreateLookAt(new Vector3(4, 6, 8), new Vector3(0, 0, 0), new Vector3(0, 1, 0));
            Matrix proj_mtx = ProjisionMatrix;
            Matrix obj_mtx;
            Vector3 obj_pos;

            // Init the rendering
            graphics.GraphicsDevice.RenderState.AlphaBlendEnable = false;
            graphics.GraphicsDevice.RenderState.SourceBlend = Blend.SourceAlpha;
            graphics.GraphicsDevice.RenderState.DestinationBlend = Blend.One;

            m_render_alpha = 1.0f;
            m_render_color = null;

            // Render the background
            RenderBackground(graphics.GraphicsDevice, ref proj_mtx);

            // Render the tiles
            for (int y = 0; y < Board.Height; ++y)
            {
                for (int x = 0; x < Board.Width; ++x)
                {
                    // Create the translation matrix
                    TilePosition(out obj_pos, x, y);
                    MakeItemMatrix(out obj_mtx, ref obj_pos, 0.0f);

                    // Render the mesh
                    RenderModel(m_res.model_tile, ref camera_mtx, ref proj_mtx, ref obj_mtx);
                }
            }

            // Render the walls
            for (int y = 0; y <= Board.Height; ++y)
            {
                for (int x = 0; x < Board.Width; ++x)
                {
                    if (m_board.wall_horiz[x, y] == false)
                        continue;

                    // Create the translation matrix
                    WallHorizPosition(out obj_pos, x, y);
                    MakeItemMatrix(out obj_mtx, ref obj_pos, 90.0f);

                    // Render the mesh
                    RenderModel(m_res.model_wall, ref camera_mtx, ref proj_mtx, ref obj_mtx);
                }
            }

            for (int y = 0; y < Board.Height; ++y)
            {
                for (int x = 0; x <= Board.Width; ++x)
                {
                    if (m_board.wall_vert[x, y] == false)
                        continue;

                    // Create the translation matrix
                    WallVertPosition(out obj_pos, x, y);
                    MakeItemMatrix(out obj_mtx, ref obj_pos, 0.0f);

                    // Render the mesh
                    RenderModel(m_res.model_wall, ref camera_mtx, ref proj_mtx, ref obj_mtx);
                }
            }

            for (int y = 0; y <= Board.Height; ++y)
            {
                for (int x = 0; x <= Board.Width; ++x)
                {
                    if (m_board.filled_intersection[x, y] == false)
                        continue;

                    // Create the translation matrix
                    WallCornerPosition(out obj_pos, x, y);
                    MakeItemMatrix(out obj_mtx, ref obj_pos, 0.0f);

                    // Render the mesh
                    RenderModel(m_res.model_wall_corner, ref camera_mtx, ref proj_mtx, ref obj_mtx);
                }
            }

            // Render the pieces
            for (int i = 0; i < m_board.pieces.Count; ++i)
            {
                if (m_board.pieces[i].Finished)
                    continue;

                Point pos = m_board.pieces[i].Position;

                m_render_color = GetCurrentPlayerColor(m_board.pieces[i].Player);

                // Create the translation matrix
                PiecePosition(out obj_pos, pos.X, pos.Y);
                MakeItemMatrix(out obj_mtx, ref obj_pos, 0.0f);

                // Render the mesh
                RenderModel(m_res.model_piece, ref camera_mtx, ref proj_mtx, ref obj_mtx);
            }
            m_render_color = null;

            // Render the players goal markers (outside the board)
            for (int i = 0; i < m_board.PlayerCount; ++i)
            {
                // Get the tile position
                Point tile_pos, tile_dir;
                int tile_len;
                PlayerPiece.GetGoalPieces((PlayerIndex)i, out tile_pos, out tile_dir, out tile_len);
                PointUtil.Add(ref tile_pos, PointUtil.Scale(tile_dir, (tile_len / 2)));

                // Get the world position
                TilePosition(out obj_pos, tile_pos.X, tile_pos.Y);
                obj_pos *= 1.3f;

                obj_pos.Y = -obj_pos.Y;
                MakeItemMatrix(out obj_mtx, ref obj_pos, 0.0f);

                obj_mtx = m_goal_marker_matrix[i] * obj_mtx;
                obj_mtx = Matrix.CreateScale(MathHelper.Clamp(1.0f - Math.Abs(obj_pos.Y) / HiddenY, 0.0f, 1.0f)) * obj_mtx;

                // Render
                m_render_color = GetCurrentPlayerColor((PlayerIndex)i);
                RenderModel(m_res.model_goal_location, ref camera_mtx, ref proj_mtx, ref obj_mtx);
                m_render_color = null;
            }

            // Render the active players goal tiles
            if (ActivePlayer != null)
            {
                graphics.GraphicsDevice.RenderState.AlphaBlendEnable = true;
                m_render_alpha = 0.1f;

                // Get the goal tiles
                Point pos, dir;
                int len;
                PlayerPiece.GetGoalPieces(ActivePlayer.Value, out pos, out dir, out len);

                while (len-- > 0)
                {
                    TilePosition(out obj_pos, pos.X, pos.Y);
                    MakeItemMatrix(out obj_mtx, ref obj_pos, 0.0f);
                    PointUtil.Add(ref pos, dir);

                    RenderModel(m_res.model_tile_goal_overlay, ref camera_mtx, ref proj_mtx, ref obj_mtx);
                }
            }

            // Render the to-be-placed piece
            if (m_showing_wall != null)
            {
                // Set up the render state
                graphics.GraphicsDevice.RenderState.AlphaBlendEnable = true;
                m_render_alpha = 0.8f;

                if (m_showing_wall.CurrentPlaceIsValid)
                    m_render_color = null;
                else
                    m_render_color = new Color(255, 0, 0);

                // Render
                RenderWallPiece(m_showing_wall, m_showing_piece_pos, m_showing_piece_ang, ref camera_mtx, ref proj_mtx);
            }

            if (m_showing_plr != null)
            {
                // Set up the render state
                graphics.GraphicsDevice.RenderState.AlphaBlendEnable = true;
                m_render_alpha = 0.5f;
                m_render_color = null;

                // Render the cursor
                obj_mtx = Matrix.CreateScale(PieceScale * 1.1f) * Matrix.CreateTranslation(m_showing_piece_pos);
                RenderModel(m_res.model_piece, ref camera_mtx, ref proj_mtx, ref obj_mtx);

                // Render the piece we're moving
                PiecePosition(out obj_pos, m_showing_plr.Position.X, m_showing_plr.Position.Y);
                MakeItemMatrix(out obj_mtx, ref obj_pos, 0.0f);
                obj_mtx = Matrix.CreateScale(1.05f) * obj_mtx;

                m_render_alpha = 0.8f;
                RenderModel(m_res.model_piece, ref camera_mtx, ref proj_mtx, ref obj_mtx);

                // Show the destination places
                m_render_alpha = 0.3f;

                List<Point> dest_lst = m_board.active_player_piece_destinations;
                for (int i = 0; i < dest_lst.Count; ++i)
                {
                    Point pos = dest_lst[i];

                    // Create the translation matrix
                    PiecePosition(out obj_pos, pos.X, pos.Y);
                    MakeItemMatrix(out obj_mtx, ref obj_pos, 0.0f);

                    // Render the marker
                    RenderModel(m_res.model_piece, ref camera_mtx, ref proj_mtx, ref obj_mtx);
                }
            }
        }

        private void RenderBackground(GraphicsDevice dev, ref Matrix proj_mtx)
        {
            Matrix camera_mtx = CameraMatrix;
            camera_mtx.Translation = Vector3.Zero;

            Matrix world_mtx = Matrix.Identity;
            //Matrix world_mtx = Matrix.CreateRotationX(MathHelper.ToRadians(180.0f)); ;

            m_render_color = null;

            dev.RenderState.DepthBufferEnable = false;
            RenderModel(m_res.model_env_cube, ref camera_mtx, ref proj_mtx, ref world_mtx);
            dev.RenderState.DepthBufferEnable = true;
        }

        private void RenderWallPiece(WallPiece piece, Vector3 pos, float angle_deg, ref Matrix camera_mtx, ref Matrix proj_mtx)
        {
            pos.X -= PieceSize * 0.5f;
            pos.Y += 0.0f;
            pos.Z -= PieceSize * 0.5f;
            Matrix scale_mtx = Matrix.CreateScale(PieceScale);
            Matrix base_mtx = Matrix.CreateScale(1.1f) * Matrix.CreateRotationY(MathHelper.ToRadians(angle_deg)) * Matrix.CreateTranslation(pos);

            for (int i = 0; i < piece.HorizWallCount; ++i)
            {
                Point board_pos = piece.GetOrigHorizWallPos(i);
                Vector3 obj_pos = new Vector3(board_pos.X * PieceSize + PieceSize * 0.5f, 0.0f, board_pos.Y * PieceSize);
                Matrix obj_mtx = scale_mtx * Matrix.CreateRotationY((float)Math.PI * 0.5f) * Matrix.CreateTranslation(obj_pos) * base_mtx;

                RenderModel(m_res.model_wall, ref camera_mtx, ref proj_mtx, ref obj_mtx);
            }

            for (int i = 0; i < piece.VertWallCount; ++i)
            {
                Point board_pos = piece.GetOrigVertWallPos(i);
                Vector3 obj_pos = new Vector3(board_pos.X * PieceSize, 0.0f, board_pos.Y * PieceSize + PieceSize * 0.5f);
                Matrix obj_mtx = scale_mtx * Matrix.CreateTranslation(obj_pos) * base_mtx;

                RenderModel(m_res.model_wall, ref camera_mtx, ref proj_mtx, ref obj_mtx);
            }

            for (int i = 0; i < piece.IntersectionCount; ++i)
            {
                Point board_pos = piece.GetOrigIntersectPos(i);
                Vector3 obj_pos = new Vector3(board_pos.X * PieceSize, 0.0f, board_pos.Y * PieceSize);
                Matrix obj_mtx = scale_mtx * Matrix.CreateTranslation(obj_pos) * base_mtx;

                RenderModel(m_res.model_wall_corner, ref camera_mtx, ref proj_mtx, ref obj_mtx);
            }
        }

        private void RenderModelMesh(ModelMesh mesh, ref Matrix camera_mtx, ref Matrix proj_mtx, ref Matrix obj_mtx)
        {
            int effect_amt = mesh.Effects.Count;
            for (int i = 0; i < effect_amt; ++i)
            {
                BasicEffect effect = (BasicEffect)mesh.Effects[i];

                effect.Projection = proj_mtx;
                effect.View = camera_mtx;
                effect.World = obj_mtx;

                effect.Alpha = m_render_alpha;
                effect.VertexColorEnabled = true;

                if (m_render_color != null)
                {
                    effect.DiffuseColor = m_render_color.Value.ToVector3();
                }
                else
                {
                    effect.DiffuseColor = new Vector3(1, 1, 1);
                }
            }

            mesh.Draw();
        }

        private void RenderModel(Model model, ref Matrix camera_mtx, ref Matrix proj_mtx, ref Matrix obj_mtx)
        {
            int mesh_amt = model.Meshes.Count;
            for (int i = 0; i < mesh_amt; ++i)
            {
                RenderModelMesh(model.Meshes[i], ref camera_mtx, ref proj_mtx, ref obj_mtx);
            }
        }

        private float GetClampedY(VelObj[,] arr, int x, int y)
        {
            if (x < 0) x = 0; else if (x >= arr.GetLength(0)) x = arr.GetLength(0) - 1;
            if (y < 0) y = 0; else if (y >= arr.GetLength(1)) y = arr.GetLength(1) - 1;
            return arr[x, y].y;
        }

        private void TilePosition(out Vector3 dest, int board_x, int board_y)
        {
            dest = new Vector3(
                CenterOffsetX + (board_x * PieceSize),
                GetClampedY(m_vis_tile, board_x, board_y),
                CenterOffsetY + (board_y * PieceSize));
        }

        private void PiecePosition(out Vector3 dest, int board_x, int board_y)
        {
            TilePosition(out dest, board_x, board_y);
            dest.Y += PieceFloorY;
        }

        private void WallHorizPosition(out Vector3 dest, int board_x, int board_y)
        {
            TilePosition(out dest, board_x, board_y);
            dest.Z -= PieceSize * 0.5f;
            dest.Y += WallFloorY + GetClampedY(m_vis_wall_horiz, board_x, board_y);
        }

        private void WallVertPosition(out Vector3 dest, int board_x, int board_y)
        {
            TilePosition(out dest, board_x, board_y);
            dest.X -= PieceSize * 0.5f;
            dest.Y += WallFloorY + GetClampedY(m_vis_wall_vert, board_x, board_y);
        }

        private void WallCornerPosition(out Vector3 dest, int board_x, int board_y)
        {
            TilePosition(out dest, board_x, board_y);
            dest.X -= PieceSize * 0.5f;
            dest.Y += WallFloorY + GetClampedY(m_vis_isect, board_x, board_y);
            dest.Z -= PieceSize * 0.5f;
        }

        private void MakeItemMatrix(out Matrix dest, ref Vector3 pos, float y_rot_deg)
        {
            Matrix scale;
            Matrix translate;

            if (y_rot_deg != 0.0f)
            {
                scale =
                    Matrix.CreateRotationY(MathHelper.ToRadians(y_rot_deg)) *
                    Matrix.CreateScale(PieceScale);
            }
            else
            {
                Matrix.CreateScale(PieceScale, out scale);
            }

            Matrix.CreateTranslation(ref pos, out translate);
            Matrix.Multiply(ref scale, ref translate, out dest);
        }

        private Color GetCurrentPlayerColor(PlayerIndex plr)
        {
            if (ActivePlayer != null && ActivePlayer.Value == plr)
            {
                float mul = 1.0f - Math.Abs((float)Math.Sin(m_color_wave_player));
                return GetPlayerColor(plr, mul * 0.4f);
            }
            else
                return GetPlayerColor(plr);
        }

        public static Color GetPlayerColor(PlayerIndex plr)
        {
            switch (plr)
            {
                default:
                case PlayerIndex.One: return Color.Red;
                case PlayerIndex.Two: return Color.Cyan;
                case PlayerIndex.Three: return Color.Yellow;
                case PlayerIndex.Four: return Color.LimeGreen;
            }
        }

        public static Color GetPlayerColor(PlayerIndex plr, float white_factor)
        {
            Color col = GetPlayerColor(plr);
            Vector3 v = Vector3.Lerp(col.ToVector3(), new Vector3(1, 1, 1), white_factor);

            return new Color(v);
        }
    }
}
