using System;
using System.Collections.Generic;
using System.Text;
using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Input;

namespace Waboga
{
    public enum ControllerButton
    {
        A,
        B,
        X,
        Y,
        PadLeft,
        PadRight,
        PadUp,
        PadDown,
        LB,
        RB,
        LT,
        RT,
        Back,
        Start,
        LeftStick,
        RightStick,

        LeftStickLeft,
        LeftStickRight,
        LeftStickUp,
        LeftStickDown,

        RightStickLeft,
        RightStickRight,
        RightStickUp,
        RightStickDown,

        MouseLeft,
        MouseRight,
        MouseWheelUp,
        MouseWheelDown,

        KeyEsc,
        KeySpace,
    }

    public class Controller
    {
        private struct MouseButtonState
        {
            public ButtonState left;
            public ButtonState right;
            public bool wheel_up;
            public bool wheel_down;
            public int wheel_value;

            private void PreventWarningsOnXbox360Build()
            {
                // Hack code to prevent warnings on the Xbox 360 build
                left = right = ButtonState.Released;
                wheel_up = wheel_down = false;
                wheel_value = 0;
            }
        }
        private struct KeyStates
        {
            public bool esc;
            public bool space;

            private void PreventWarningsOnXbox360Build()
            {
                // Hack code to prevent warnings on the Xbox 360 build
                esc = space = false;
            }
        }

        private static Vector2 DragCursorPosition = new Vector2(100, 100);

        private GamePadState[] m_last_state = new GamePadState[4];
        private GamePadState[] m_cur_state = new GamePadState[4];
        private Vector2 m_left_stick = Vector2.Zero;
        private Vector2 m_right_stick = Vector2.Zero;
        private bool m_mouse_drag_mode = false;
        private Vector2 m_mouse_pos = Vector2.Zero;
        private Vector2 m_mouse_drag = Vector2.Zero;
        private Vector2 m_mouse_scale = new Vector2(1.0f, 1.0f);
        private MouseButtonState m_mouse_buttons_last;
        private MouseButtonState m_mouse_buttons;
        private KeyStates m_keys_last;
        private KeyStates m_keys;

        public Vector2 LeftStick
        {
            get { return m_left_stick; }
        }

        public Vector2 RightStick
        {
            get { return m_right_stick; }
        }

        public Vector2 MousePos
        {
            get { return m_mouse_pos * m_mouse_scale; }
        }

        public Vector2 UnscaledMousePos
        {
            get { return m_mouse_pos; }
        }

        public Vector2 MousePosScale
        {
            get { return m_mouse_scale; }
            set { m_mouse_scale = value; }
        }

        public bool MouseDragMode
        {
            get { return m_mouse_drag_mode; }
            set
            {
                if (m_mouse_drag_mode == value)
                    return;
                m_mouse_drag_mode = value;
                m_mouse_drag = Vector2.Zero;

                if (m_mouse_drag_mode)
                { // Entering drag mode
#if !XBOX
                    if (WindowActive)
                        Mouse.SetPosition((int)DragCursorPosition.X, (int)DragCursorPosition.Y);
#endif
                }
                else
                { // Leaving drag mode
#if !XBOX
                    if (WindowActive)
                        Mouse.SetPosition((int)m_mouse_pos.X, (int)m_mouse_pos.Y);
#endif
                }
            }
        }

        public Vector2 MouseDragAmount
        {
            get { return m_mouse_drag; }
        }

        private bool m_window_active;
        public bool WindowActive
        {
            get { return m_window_active; }
            set { m_window_active = value; }
        }



        public Controller()
        {
            for (int i = 0; i < 4; ++i)
            {
                m_last_state[i] = new GamePadState();
                m_cur_state[i] = new GamePadState();
            }
        }

        public void Update()
        {
            // Get the mouse state
#if !XBOX
            MouseState mouse = Mouse.GetState();
            Vector2 mouse_pos = new Vector2((float)mouse.X, (float)mouse.Y);
            if (MouseDragMode)
            {
                m_mouse_drag = mouse_pos - DragCursorPosition;
                if (WindowActive)
                    Mouse.SetPosition((int)DragCursorPosition.X, (int)DragCursorPosition.Y);
            }
            else
            {
                m_mouse_pos = mouse_pos;
                m_mouse_drag = Vector2.Zero;
            }


            int wheel_value = m_mouse_buttons_last.wheel_value;
            m_mouse_buttons_last = m_mouse_buttons;

            if (WindowActive)
            {
                m_mouse_buttons.left = mouse.LeftButton;
                m_mouse_buttons.right = mouse.RightButton;
                m_mouse_buttons.wheel_up = (mouse.ScrollWheelValue < wheel_value);
                m_mouse_buttons.wheel_down = (mouse.ScrollWheelValue > wheel_value);
                m_mouse_buttons.wheel_value = mouse.ScrollWheelValue;
            }
            else
            {
                m_mouse_buttons.left = ButtonState.Released;
                m_mouse_buttons.right = ButtonState.Released;
                m_mouse_buttons.wheel_up = false;
                m_mouse_buttons.wheel_down = false;
                m_mouse_buttons.wheel_value = mouse.ScrollWheelValue;
            }
#endif

            // Get the keyboard state
#if !XBOX
            KeyboardState keys = Keyboard.GetState();
            m_keys_last = m_keys;
            m_keys.esc = WindowActive ? keys.IsKeyDown(Keys.Escape) : false;
            m_keys.space = WindowActive ? keys.IsKeyDown(Keys.Space) : false;
#endif

            // Read the controllers
            for (int i = 0; i < 4; ++i)
            {
                m_last_state[i] = m_cur_state[i];
                m_cur_state[i] = GamePad.GetState((PlayerIndex)i);

                // Update sticks (get the most dominant one)
                Vector2 lstick = m_cur_state[i].ThumbSticks.Left;
                Vector2 rstick = m_cur_state[i].ThumbSticks.Right;

                if (i == 0)
                {
                    m_left_stick = lstick;
                    m_right_stick = rstick;
                }
                else
                {
                    if (lstick.LengthSquared() > m_left_stick.LengthSquared())
                        m_left_stick = lstick;

                    if (rstick.LengthSquared() > m_right_stick.LengthSquared())
                        m_right_stick = rstick;
                }
            }
        }

        public bool IsDown(ControllerButton button)
        {
            for (int i = 0; i < 4; ++i)
            {
                if (GetButton(ref m_cur_state[i], ref m_mouse_buttons, ref m_keys, button))
                    return true;
            }
            return false;
        }

        public bool IsUp(ControllerButton button)
        {
            return !IsDown(button);
        }

        public bool WasDown(ControllerButton button)
        {
            for (int i = 0; i < 4; ++i)
            {
                if (GetButton(ref m_last_state[i], ref m_mouse_buttons_last, ref m_keys_last, button))
                    return true;
            }
            return false;
        }

        public bool WasUp(ControllerButton button)
        {
            return !WasDown(button);
        }

        public bool WasPressed(ControllerButton button)
        {
            return IsDown(button) && WasUp(button);
        }

        public bool WasReleased(ControllerButton button)
        {
            return IsUp(button) && WasDown(button);
        }

        private bool GetButton(ref GamePadState state, ref MouseButtonState mb_state, ref KeyStates key_state, ControllerButton button)
        {
            switch (button)
            {
                default: return false;
                case ControllerButton.A: return (state.Buttons.A == ButtonState.Pressed);
                case ControllerButton.B: return (state.Buttons.B == ButtonState.Pressed);
                case ControllerButton.X: return (state.Buttons.X == ButtonState.Pressed);
                case ControllerButton.Y: return (state.Buttons.Y == ButtonState.Pressed);
                case ControllerButton.PadLeft: return (state.DPad.Left == ButtonState.Pressed);
                case ControllerButton.PadRight: return (state.DPad.Right == ButtonState.Pressed);
                case ControllerButton.PadUp: return (state.DPad.Up == ButtonState.Pressed);
                case ControllerButton.PadDown: return (state.DPad.Down == ButtonState.Pressed);
                case ControllerButton.LB: return (state.Buttons.LeftShoulder == ButtonState.Pressed);
                case ControllerButton.RB: return (state.Buttons.RightShoulder == ButtonState.Pressed);
                case ControllerButton.LT: return (state.Triggers.Left >= 0.5f);
                case ControllerButton.RT: return (state.Triggers.Right >= 0.5f);
                case ControllerButton.Back: return (state.Buttons.Back == ButtonState.Pressed);
                case ControllerButton.Start: return (state.Buttons.Start == ButtonState.Pressed);
                case ControllerButton.LeftStick: return (state.Buttons.LeftStick == ButtonState.Pressed);
                case ControllerButton.RightStick: return (state.Buttons.RightStick == ButtonState.Pressed);
                case ControllerButton.LeftStickLeft: return (state.ThumbSticks.Left.X <= -0.5f);
                case ControllerButton.LeftStickRight: return (state.ThumbSticks.Left.X >= 0.5f);
                case ControllerButton.LeftStickUp: return (state.ThumbSticks.Left.Y >= 0.5f);
                case ControllerButton.LeftStickDown: return (state.ThumbSticks.Left.Y <= -0.5f);
                case ControllerButton.RightStickLeft: return (state.ThumbSticks.Right.X <= -0.5f);
                case ControllerButton.RightStickRight: return (state.ThumbSticks.Right.X >= 0.5f);
                case ControllerButton.RightStickUp: return (state.ThumbSticks.Right.Y >= 0.5f);
                case ControllerButton.RightStickDown: return (state.ThumbSticks.Right.Y <= -0.5f);
                case ControllerButton.MouseLeft: return (mb_state.left == ButtonState.Pressed);
                case ControllerButton.MouseRight: return (mb_state.right == ButtonState.Pressed);
                case ControllerButton.MouseWheelUp: return mb_state.wheel_up;
                case ControllerButton.MouseWheelDown: return mb_state.wheel_down;
                case ControllerButton.KeyEsc: return key_state.esc;
                case ControllerButton.KeySpace: return key_state.space;
            }
        }
    }
}
