using System;
using System.Collections.Generic;
using System.Text;
using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Graphics;
using Microsoft.Xna.Framework.Graphics.PackedVector;

namespace Waboga
{
    public class TexturePainter
    {
        private GraphicsDevice m_device;
        private int m_width;
        private int m_height;
        private bool m_do_mipmaps;
        private Color[] m_data;

        public Texture2D Texture
        {
            get
            {
                Texture2D tex = new Texture2D(
                    m_device,
                    m_width,
                    m_height,
                    m_do_mipmaps ? 0 : 1,
                    ResourceUsage.Tiled,
                    SurfaceFormat.Color);
                tex.SetData<Color>(m_data);
                if (m_do_mipmaps)
                    tex.GenerateMipMaps(TextureFilter.Linear);
                return tex;
            }
        }

        public int Width
        {
            get { return m_width; }
        }

        public int Height
        {
            get { return m_height; }
        }

        public TexturePainter(GraphicsDevice dev, int w, int h, bool do_mipmaps)
        {
            m_device = dev;
            m_width = w;
            m_height = h;
            m_do_mipmaps = do_mipmaps;
            m_data = new Color[m_width * m_height];

            Clear(Color.Black);
        }

        public void Clear(Color c)
        {
            for (int i = 0; i < m_data.Length; ++i)
                m_data[i] = Color.Black;
        }

        public void MakeColorTransparent(Color c)
        {
            Color transp = new Color(0, 0, 0, 0);
            for (int i = 0; i < m_data.Length; ++i)
            {
                if (m_data[i] == c)
                    m_data[i] = transp;
            }
        }

        public void FillRect(Color c, Rectangle rect)
        {
            FillRect(c, rect.X, rect.Y, rect.Width, rect.Height);
        }

        public void FillRect(Color c, int x, int y, int w, int h)
        {
            // Make sure the width/height is not below zero on input
            if (w < 0)
            {
                w = -w;
                x -= w;
            }
            if (h < 0)
            {
                h = -h;
                y -= h;
            }

            // Clip the rectangle
            if (x < 0)
            {
                w -= (-x);
                x = 0;
            }
            if (y < 0)
            {
                h -= (-y);
                y = 0;
            }
            if ((x + w) > m_width)
                w -= m_width - (x + w);
            if ((y + h) > m_height)
                w -= m_height - (y + h);

            if (w <= 0 || h <= 0)
                return;

            // Draw
            while (h-- > 0)
            {
                int idx = x + y * m_width;
                ++y;

                int xleft = w;
                while (xleft-- > 0)
                    m_data[idx++] = c;
            }
        }

        public void Scroll(int xd, int yd)
        {
            if (xd == 0 && yd == 0)
                return;

            int x_read = MathUtil.WrapRange(-xd, 0, Width);
            int y_read = MathUtil.WrapRange(-yd, 0, Height);

            // Naive memory/GC wasting implementation
            Color[] old_data = m_data;
            m_data = new Color[old_data.Length];

            int didx = 0;
            for (int y = 0; y < m_height; ++y)
            {
                int sidx = y_read * m_width;
                int sx = x_read;

                for (int x = 0; x < m_width; ++x)
                {
                    m_data[didx++] = old_data[sidx + sx];
                    sx = (sx + 1) % m_width;
                }

                y_read = (y_read + 1) % m_height;
            }
        }

        public Color GetClamped(int x, int y)
        {
            if (x < 0) x = 0;
            if (y < 0) y = 0;
            if (x >= m_width) x = m_width - 1;
            if (y >= m_height) y = m_height - 1;

            return m_data[x + y * m_width];
        }

        public void SurroundAlphaByColor(Color c)
        {
            Color[] orig_data = m_data;
            m_data = new Color[orig_data.Length];

            int w1 = m_width - 1;
            int h1 = m_height - 1;

            int idx = 0;
            for (int y = 0; y < m_height; ++y)
            {
                for (int x = 0; x < m_width; ++x, ++idx)
                {
                    if (orig_data[idx].A > 0)
                    {
                        m_data[idx] = orig_data[idx];
                        continue;
                    }

                    bool has_adjacent_data = false;
                    if (x > 0 && orig_data[idx - 1].A > 0)
                        has_adjacent_data = true;
                    else if (x < w1 && orig_data[idx + 1].A > 0)
                        has_adjacent_data = true;
                    else if (y > 0 && orig_data[idx - m_width].A > 0)
                        has_adjacent_data = true;
                    else if (y < h1 && orig_data[idx + m_width].A > 0)
                        has_adjacent_data = true;

                    if (has_adjacent_data)
                        m_data[idx] = c;
                    else
                        m_data[idx] = orig_data[idx];
                }
            }
        }
    }
}
