using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.Text;
using System.Threading;
using Waboga;
using Microsoft.Xna.Framework;

namespace Waboga.UI
{
    public struct GameAIMove
    {
        public PlayerIndex player;
        public bool wall_instead_of_piece;
        public int index; // Wall or piece index
        public int angle; // Wall angle
        public Point position; // The position to place at/move to
    }

    public class GameAI
    {
        private const bool UseTimeInsteadOfDepth = false;
        private const int MaxDepth = 1;
        private const int MaxSecs = 5;
        private const int MaxBranchesPerLevel = 9999;
        private const int MaxMovesPerWall = 9999;

        private Board m_play_board;
        private Board m_ai_board;

        private Thread m_running_thread;
        private GameAIMove? m_final_move;

        private List<Point> m_temp_points = new List<Point>();
        private GameAIStatePool m_state_pool;
        private PlayerIndex m_start_plr;
        private PlayerInventory[] m_start_inventory;

        private bool[,] m_temp_can_test = new bool[Board.Width + 1, Board.Height + 1];

        public GameAI(Board b)
        {
            m_play_board = b;
            m_ai_board = new Board();
            m_ai_board.Init(b.PlayerCount, b.PieceCount);

            m_state_pool = new GameAIStatePool(b.PlayerCount);
        }

        public void StartAI(PlayerIndex plr, PlayerInventory[] inv)
        {
            Debug.Assert(m_play_board.PlayerCount == 2);

            lock (this)
            {
                if (m_running_thread != null)
                    throw new Exception("Already running!");

                m_final_move = null;
                m_start_plr = plr;
                m_start_inventory = inv;

                m_running_thread = new Thread(MainAI);
#if XBOX
                m_running_thread.SetProcessorAffinity(4);
#endif
                m_running_thread.Start();
            }
        }

        public void ForceQuit()
        {
            lock (this)
            {
                if (m_running_thread != null)
                {
                    try
                    {
                        m_running_thread.Abort();
                    }
                    catch
                    {
                    }
                }
            }
        }

        public bool AIDone()
        {
            lock (this)
            {
                return (m_running_thread == null);
            }
        }

        public void DoFinalMove()
        {
            lock (this)
            {
                Debug.Assert(AIDone());

                // Apply the final move to the board
                if (m_final_move != null)
                {
                    ApplyMove(m_final_move.Value, m_play_board);

                    // Deduct a wall piece?
                    if (m_final_move.Value.wall_instead_of_piece)
                    {
                        m_start_inventory[(int)m_start_plr].PieceCount[m_final_move.Value.index] -= 1;
                    }
                }

                m_final_move = null;
            }
        }

        private void MainAI()
        {
            // Get the initial state
            m_state_pool.Clear();
            GameAIState state = m_state_pool.Alloc();

            state.plr = m_start_plr;
            state.CopyInventory(m_start_inventory);
            state.board_state.CopyFrom(m_play_board);

            /*state.board_state.CopyTo(m_ai_board);
            Console.WriteLine("AI dist: " + DistToNextLine(state.plr));
            Console.WriteLine("Plr dist: " + DistToNextLine(state.NextPlayer));*/

            m_final_move = null;
            if (UseTimeInsteadOfDepth)
                IterativeAlphaBetaNegamax(state);
            else
                AlphaBetaNegamax(state, MaxDepth);

            // Done
            lock (this)
            {
                m_running_thread = null;
            }
        }

        private int m_current_depth;
        private int m_last_run_successor_count;

        private float IterativeAlphaBetaNegamax(GameAIState state)
        {
            float ret = 0.0f;

            m_current_depth = 0;

            int depth = 0;
            DateTime start_time = DateTime.Now;
            TimeSpan span;
            while (true)
            {
                span = DateTime.Now - start_time;
                if (span.Seconds >= MaxSecs)
                    break;

                // Test it
                ++depth;
                ret = AlphaBetaNegamax(state, depth);

                // If we got way too many successors the last run, do not go too deep
                /*if (m_last_run_successor_count > (500 / depth))
                {
                    Console.WriteLine("AI: Successor count too high; preventing further work");
                    break;
                }*/
            }

            span = DateTime.Now - start_time;
            Console.WriteLine("AI: Got to iteration depth " + m_current_depth + " in " + span.Seconds + " seconds");

            return ret;
        }

        private float AlphaBetaNegamax(GameAIState state, int depth)
        {
            m_current_depth = depth;
            return NegamaxValue(state, depth, -9999.0f, 9999.0f);
        }

        private float NegamaxValue(GameAIState state, int depth, float alpha, float beta)
        {
            // Do the terminal test and return the utility value
            state.board_state.CopyTo(m_ai_board);

            if (m_ai_board.HasGameCompleted)
            {
                if (m_ai_board.HasPlayerCompleted(state.plr))
                    return 100.0f;
                else
                    return -100.0f;
            }

            // If at maximum recursion depth, evaluate the state
            if (depth <= 0)
                return EvaluateCurrent(state);

            // Get the list of successors
            List<GameAIMove> successors = state.successors;
            successors.Clear();

            GetSuccessors(successors, state);

            if (depth == m_current_depth)
            {
                m_last_run_successor_count = successors.Count;
                Console.WriteLine("AI: Toplevel has " + m_last_run_successor_count + " branches (max search depth is " + m_current_depth + ")");
            }

            // Prepare the new state
            GameAIState new_state = m_state_pool.Alloc();
            new_state.plr = state.NextPlayer;

            // Find the best move
            float v = -9999.0f;

            for (int midx = 0; midx < successors.Count; ++midx)
            {
                GameAIMove do_move = successors[midx];

                // Apply the move
                new_state.CopyInventory(state);

                state.board_state.CopyTo(m_ai_board);
                ApplyMove(do_move, m_ai_board);
                new_state.board_state.CopyFrom(m_ai_board);

                if (do_move.wall_instead_of_piece)
                { // Update the player inventory
                    new_state.plr_inv[(int)new_state.plr][do_move.index] -= 1;
                }

                // Test the child node
                float child = -NegamaxValue(new_state, depth - 1, -beta, -alpha);

                if (child > v)
                {
                    if (depth == m_current_depth)
                        m_final_move = do_move;
                    v = child;

                    // Keep track of the alpha value
                    if (v > alpha)
                    {
                        alpha = v;

                        // Prune
                        if (alpha >= beta)
                            break;
                    }
                }
            }

            if (depth == m_current_depth)
            {
                Console.WriteLine("AI: Got a move, heurestic value " + v);
            }

            // Done
            m_state_pool.Free(new_state);
            return v;
        }

        private float EvaluateCurrent(GameAIState state)
        {
// *** TODO ***
            float feat1 = PositionFeature(state.plr);
            float feat2 = -PositionFeature(state.NextPlayer);
            float feat3 = DifferenceFeature(state.plr, state.NextPlayer);
            float feat4 = DistToNextLine(state.plr);
            float feat5 = -DistToNextLine(state.NextPlayer);
            //float feat4 = 0.0f;
            //float feat5 = 0.0f;

            return
                2.0f * feat1 +
                1.0f * feat2 +
                0.0f * feat3 +
                2.0f * feat4 +
                5.0f * feat5;
        }

        private float PositionFeature(PlayerIndex plr)
        {
            float best = 0.0f;

            int pieces = m_ai_board.PieceCount;
            for (int pidx = 0; pidx < pieces; ++pidx)
            {
                PlayerPiece piece = m_ai_board.GetPiece(plr, pidx);
                if (piece.Finished)
                    continue;

                // Distance from goal line
                float dist;
                if (plr == PlayerIndex.One)
                    dist = piece.Position.Y;
                else
                    dist = (Board.Height - 1) - piece.Position.Y;

                // Get the comparison value
                float val = 1.0f - dist / (float)(Board.Height - 1);
                if (val > best)
                    best = val;
            }

            return best;
        }

        private float DifferenceFeature(PlayerIndex plr, PlayerIndex plr_opp)
        {
            float val1 = PositionFeature(plr);
            float val2 = PositionFeature(plr_opp);

            return val1 - val2;
        }

        private float DistToNextLine(PlayerIndex plr)
        {
            int shortest = 99999;

            int pieces = m_ai_board.PieceCount;
            for (int pidx = 0; pidx < pieces; ++pidx)
            {
                PlayerPiece piece = m_ai_board.GetPiece(plr, pidx);
                if (piece.Finished)
                    continue;

                // Distance from goal line
                int line = piece.Position.Y;
                if (plr == PlayerIndex.One)
                    line -= 1;
                else
                    line += 1;

                // Get the distance
                int dist = m_ai_board.DistanceToHorizLine(line, piece);
                if (dist < shortest)
                    shortest = dist;
            }

            //Console.WriteLine("SHORTEST: " + shortest + " " + plr);

            float goodness = (float)shortest * (1.0f / (float)(Board.Width * Board.Height));
            return 1.0f - goodness;
        }

        private void ApplyMove(GameAIMove move, Board board)
        {
            if (move.wall_instead_of_piece)
            {
                WallPiece piece = GamePage.WallPieces[move.index];
                piece.Angle = move.angle;
                piece.Position = move.position;
                piece.PlaceOnBoard(board);
            }
            else
            {
                PlayerPiece piece = board.GetPiece(move.player, move.index);
                piece.Position = move.position;
            }
        }

        private RandomSkipper m_skipper_piece = new RandomSkipper();
        private RandomSkipper m_skipper_xy = new RandomSkipper();

        private void GetSuccessors(List<GameAIMove> dest, GameAIState state)
        {
            Array.Clear(m_temp_can_test, 0, m_temp_can_test.Length);

            // Try moving
            for (int i = 0; i < m_ai_board.PieceCount; ++i)
            {
                PlayerPiece button = m_ai_board.GetPiece(state.plr, i);
                if (button.Finished)
                    continue;

                m_ai_board.GetPieceDestinations(button, m_temp_points);
                m_skipper_piece.Init(m_temp_points.Count);
                while (!m_skipper_piece.Done)
                {
                    GameAIMove move = new GameAIMove();
                    move.player = state.plr;
                    move.index = i;
                    move.position = m_temp_points[m_skipper_piece.Next()];
                    dest.Add(move);

                    FillCanTest(move.position.X, move.position.Y, 4, 4);
                }
            }

            // Fill every wall on the board
            for (int y = 0; y <= Board.Height; ++y)
            {
                for (int x = 0; x <= Board.Width; ++x)
                {
                    if (x < Board.Width && m_ai_board.wall_horiz[x, y])
                        FillCanTest(x, y, 4, 4);
                    if (y < Board.Height && m_ai_board.wall_vert[x, y])
                        FillCanTest(x, y, 4, 4);
                }
            }

            // Try every piece where it fits
            m_skipper_piece.Init(GamePage.WallPieces.Length);
            while (!m_skipper_piece.Done)
            {
                int pidx = m_skipper_piece.Next();
                WallPiece piece = GamePage.WallPieces[pidx];
                int angles = piece.UniqueAngles;

                if (state.plr_inv[(int)state.plr][pidx] <= 0)
                    continue;

                int wall_moves = 0;

                int w = (Board.Width + 1);
                int h = (Board.Height + 1);
                m_skipper_xy.Init(w * h * angles);
                while (!m_skipper_xy.Done)
                {
                    int pos_angle_index = m_skipper_xy.Next();

                    piece.Angle = pos_angle_index % angles;
                    pos_angle_index /= angles;

                    piece.Position.X = pos_angle_index % w;
                    piece.Position.Y = pos_angle_index / w;

                    // Valid to place?
                    if (m_temp_can_test[piece.Position.X, piece.Position.Y] == false)
                        continue;
                    if (!piece.ValidOnBoard(m_ai_board))
                        continue;

                    // Add a move
                    GameAIMove move = new GameAIMove();
                    move.player = state.plr;
                    move.wall_instead_of_piece = true;
                    move.index = pidx;
                    move.position = piece.Position;
                    dest.Add(move);

                    if (dest.Count >= MaxBranchesPerLevel)
                        return;

                    ++wall_moves;
                    if (wall_moves >= MaxMovesPerWall)
                        break;
                }
            }
        }

        private void FillCanTest(int x, int y, int w, int h)
        {
            x -= (w+1) / 2;
            y -= (h+1) / 2;

            int bw = m_temp_can_test.GetLength(0);
            int bh = m_temp_can_test.GetLength(0);

            while (h-- > 0)
            {
                if (y >= bh)
                    break;

                if (y >= 0)
                {
                    int sx = x;
                    int xl = w;
                    while (xl-- > 0)
                    {
                        if (sx >= bw)
                            break;
                        if (sx >= 0)
                        {
                            m_temp_can_test[sx, y] = true;
                        }
                        ++sx;
                    }
                }

                ++y;
            }
        }
    }
}
