using System;
using System.Collections.Generic;
using System.Text;
using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Graphics;
using Microsoft.Xna.Framework.Input;
using Waboga;

namespace Waboga.UI
{
    public enum GamePieces
    {
        PieceStraight1,
        PieceStraight2,
        PieceStraight3,
        PieceCorner,
        PieceCross,
        PieceS1,
        PieceS2,
        PieceL1,
        PieceL2,
        Button1,
        Button2,
        Button3,
        Button4,
        Button5,
        Button6,
        Button7,
        Button8,
        Button9,
        __LAST__
    }

    public class PlayerInventory
    {
        public PlayerIndex Index;
        public int[] PieceCount;

        public GamePieces ActivePiece = GamePieces.Button1;
        public int PieceAngle;
        public Point PiecePosition;

        public int FinishPosition;

        public PlayerInventory(PlayerIndex index, GameStartParams start_params)
        {
            Index = index;
            PieceCount = new int[(int)GamePieces.Button1];
            PieceAngle = 0;
            PiecePosition.X = Board.Width / 2;
            PiecePosition.Y = Board.Height / 2;
            FinishPosition = -1;

            for (int i = 0; i < PieceCount.Length; ++i)
            {
                PieceCount[i] = start_params.piece_counts[i];
            }
        }
    }

    public class GamePage : UIPage
    {
        private enum AITurnState
        {
            None,
            WaitingForAI,
            AllDone,
        }

        private static string[] PlayerStrings =
            {
                "Player 1",
                "Player 2",
                "Player 3",
                "Player 4",
            };

        // Start of GamePieces -list
        public static WallPiece[] WallPieces =
            {
                WallPiece.PieceStraight1,
                WallPiece.PieceStraight2,
                WallPiece.PieceStraight3,
                WallPiece.PieceCorner,
                WallPiece.PieceCross,
                WallPiece.PieceS1,
                WallPiece.PieceS2,
                WallPiece.PieceL1,
                WallPiece.PieceL2,
            };

        private const float ShowActivePlayerTime = 1.5f;
        private const float ShowActivePlayerFadeInTime = 0.3f;
        private const float ShowActivePlayerFadeOutTime = 0.3f;

        private GameStartParams m_start_params;
        private PlayerInventory[] m_players;
        private int m_active_player;

        private GameAI m_ai;
        private AITurnState m_ai_state = AITurnState.None;

        private float m_move_pause = 0.0f;

        private int m_finished_count;
        private bool m_game_complete;

        private float m_show_active_player_countdown;

        private PlayerInventory ActivePlayer
        {
            get { return m_players[m_active_player]; }
        }

        public GamePage(GameStartParams start_params)
        {
            m_start_params = start_params;
        }

        public override void Init()
        {
            int plr_count = m_start_params.players;
            int piece_count = m_start_params.pieces_per_player;

            // Init the board
            m_finished_count = 0;
            m_game_complete = false;
            m_game_res.board.Init(plr_count, piece_count);

            // Show the board
            m_game_res.board_vis.BoardVisible = true;

            // Init the set of players
            m_active_player = 0;
            m_players = new PlayerInventory[plr_count];
            for (int i = 0; i < m_players.Length; ++i)
                m_players[i] = new PlayerInventory((PlayerIndex)i, m_start_params);

            // Init the AI
            if (m_start_params.has_computer_ai)
                m_ai = new GameAI(m_game_res.board);
            else
                m_ai = null;

            // Init this turn
            InitTurn();
        }

        public override void Deinit()
        {
            m_game_res.board_vis.BoardVisible = false;

            if (m_ai != null)
                m_ai.ForceQuit();
        }

        public override void Update(MainUI ui, float delta_t)
        {
            if (!IsTopmost)
                return;

            // Move to pause?
            if (Controller.WasPressed(ControllerButton.Start) ||
                Controller.WasPressed(ControllerButton.KeyEsc))
            {
                m_res.sound_bank.PlayCue("MenuSelect");
                ui.PushPage(new PausePage());
                return;
            }

            // Update the game
            if (m_ai != null && m_active_player == 1)
                UpdateAI();
            else
                UpdateGame(delta_t);

            // Update the "show active player" timer
            m_show_active_player_countdown -= delta_t;
            if (m_show_active_player_countdown < 0.0f)
                m_show_active_player_countdown = 0.0f;

            // Update the UI on top
            UpdateUI();

            // Game complete?
            if (m_game_complete)
                ui.PushPage(new GameDonePage());
        }

        private void UpdateAI()
        {
            m_game_res.board.active_edit_piece = null;
            m_game_res.board.active_player_piece = null;

            switch (m_ai_state)
            {
                case AITurnState.None:
                    if (m_game_res.board.HasGameCompleted ||
                        m_game_res.board.HasPlayerCompleted((PlayerIndex)m_active_player))
                    {
                        m_ai_state = AITurnState.AllDone;
                        return;
                    }

                    m_ai.StartAI((PlayerIndex)m_active_player, m_players);
                    m_ai_state = AITurnState.WaitingForAI;
                    break;

                case AITurnState.WaitingForAI:
                    if (!m_ai.AIDone())
                        return;
                    m_ai.DoFinalMove();
                    m_ai_state = AITurnState.None;

                    NextTurn();
                    break;

                case AITurnState.AllDone:
                    break;
            }
        }

        private void UpdateGame(float delta_t)
        {
            Board board = m_game_res.board;
            BoardVisualizer board_vis = m_game_res.board_vis;

            // Piece moving by a mouse
            if (!m_res.controller.MouseDragMode)
            {
                bool wall_mode = (board.active_edit_piece != null);
                Point? mouse_pos = GetMouseBoardLocation(wall_mode);

                if (mouse_pos != null)
                    ActivePlayer.PiecePosition = mouse_pos.Value;
                else
                { // Clicking outside board?
                    if (m_res.controller.WasPressed(ControllerButton.MouseLeft))
                    {
                        m_res.controller.MouseDragMode = true;
                        return;
                    }
                }
            }
            else
            {
                if (m_res.controller.IsUp(ControllerButton.MouseLeft))
                    m_res.controller.MouseDragMode = false;
                return;
            }

            // Piece moving
            Vector2 move_stick = m_res.controller.LeftStick;
            Point board_move = board_vis.GetBoardDirectionFromController(move_stick);
            if (board_move != Point.Zero)
            {
                m_move_pause -= delta_t;
                if (m_move_pause <= 0.0f)
                {
                    PointUtil.Add(ref ActivePlayer.PiecePosition, board_move);
                    m_move_pause = 0.2f;
                }
            }
            else
                m_move_pause = 0.0f;

            // Piece placing
            if (m_res.controller.WasPressed(ControllerButton.A) ||
                m_res.controller.WasPressed(ControllerButton.MouseLeft))
            {
                if (MakeMove())
                {
                    NextTurn();
                    return;
                }
            }

            // Piece turning
            if (m_res.controller.WasPressed(ControllerButton.LT))
            {
                m_res.sound_bank.PlayCue("MenuSelect");
                ActivePlayer.PieceAngle = MathUtil.WrapRange(ActivePlayer.PieceAngle - 1, 0, 4);
            }

            if (m_res.controller.WasPressed(ControllerButton.RT) ||
                m_res.controller.WasPressed(ControllerButton.MouseRight))
            {
                m_res.sound_bank.PlayCue("MenuSelect");
                ActivePlayer.PieceAngle = MathUtil.WrapRange(ActivePlayer.PieceAngle + 1, 0, 4);
            }

            // Piece changing
            bool change_left =
                m_res.controller.WasPressed(ControllerButton.LB) ||
                m_res.controller.WasPressed(ControllerButton.MouseWheelUp);
            bool change_right = 
                m_res.controller.WasPressed(ControllerButton.RB) ||
                m_res.controller.WasPressed(ControllerButton.MouseWheelDown) ||
                m_res.controller.WasPressed(ControllerButton.KeySpace);
            if (change_left || change_right)
            {
                if (change_left)
                {
                    m_res.sound_bank.PlayCue("MenuSelect");
                    MoveActiveToValidPiece(-1, -1);
                }
                if (change_right)
                {
                    m_res.sound_bank.PlayCue("MenuSelect");
                    MoveActiveToValidPiece(1, 1);
                }
            }

            // Validate the position
            ValidatePosition();

            // Update the boards active piece/position
            if (board.active_edit_piece != null)
            {
                if (board.active_edit_piece.Position != ActivePlayer.PiecePosition ||
                    board.active_edit_piece.Angle != ActivePlayer.PieceAngle)
                {
                    board.active_edit_piece.Position = ActivePlayer.PiecePosition;
                    board.active_edit_piece.Angle = ActivePlayer.PieceAngle;

                    // Test if the current piece position is valid
                    m_game_res.board.active_edit_piece.CurrentPlaceIsValid =
                        m_game_res.board.active_edit_piece.ValidOnBoard(m_game_res.board);
                }
            }

            if (board.active_player_piece != null)
            {
                board.active_player_piece_position = ActivePlayer.PiecePosition;
            }
        }

        private void UpdateUI()
        {
            Vector2 infobox_start = new Vector2(30, 30);
            Vector2 infobox_size = new Vector2(140, 120);
            Vector2 infobox_next = new Vector2(0, infobox_size.Y + 5);

            Vector2 infobox_pos = infobox_start;

            UIRenderer renderer = UIBegin();

            // Show the active player
            if (m_show_active_player_countdown > 0.0f)
            {
                // Get the text name + fade in/out
                Color plr_col = BoardVisualizer.GetPlayerColor((PlayerIndex)m_active_player);
                float alpha = 1.0f;
                if (m_show_active_player_countdown < ShowActivePlayerFadeOutTime)
                    alpha = m_show_active_player_countdown / ShowActivePlayerFadeOutTime;
                if (m_show_active_player_countdown > (ShowActivePlayerTime - ShowActivePlayerFadeInTime))
                    alpha = (ShowActivePlayerTime - m_show_active_player_countdown) / ShowActivePlayerFadeInTime;
                Color final_col = new Color(plr_col.R, plr_col.G, plr_col.B, (byte)(alpha * 255.0f));

                // Write the text
                renderer.WindowStartCentered(40.0f, 800.0f, 200.0f, new Color(0, 0, 0, 0));
                renderer.WriteLineCentered(PlayerStrings[m_active_player], final_col, 4.0f);
                renderer.WindowEnd();
            }

            // Show the player infos
            for (int plri = 0; plri < m_game_res.board.PlayerCount; ++plri)
            {
                byte color = (byte) ((plri == m_active_player) ? 150 : 20);

                renderer.WindowStart(infobox_pos, infobox_pos + infobox_size, new Color(0, 0, 0, color));
                infobox_pos += infobox_next;

                // Write the player name
                renderer.WriteLineCentered(PlayerStrings[plri], BoardVisualizer.GetPlayerColor((PlayerIndex)plri, 0.75f));

                if (m_players[plri].FinishPosition < 0)
                {
                    // Show the piece counts
                    for (int i = 0; i < (int)GamePieces.Button1; ++i)
                    {
                        int col = i % 3;
                        int row = i / 3;
                        int dest_x = 5 + col * 45;
                        int dest_y = 35 + row * 25;

                        Color text_col = new Color(0x77, 0x77, 0x77);
                        if (plri == m_active_player && i == (int)ActivePlayer.ActivePiece)
                            text_col = Color.White;

                        renderer.DrawIconAndTextAt(
                            new Vector2(dest_x, dest_y),
                            m_res.tex_wall_icons[i],
                            m_players[plri].PieceCount[i].ToString(),
                            text_col,
                            0.6f);
                    }
                }
                else
                {
                    // Draw the finished trophy
                    renderer.DrawIconAtCentered(m_res.tex_trophies[m_players[plri].FinishPosition], 35, 0.15f);
                }

                renderer.WindowEnd();
            }

            UIEnd();
        }

        private void InitTurn()
        {
            // Set the active piece selection to a valid selection
            MoveActiveToValidPiece(1, 0);

            // Update the board visualization
            m_game_res.board_vis.ActivePlayer = ActivePlayer.Index;

            // Show the player in turn
            m_show_active_player_countdown = ShowActivePlayerTime;
        }

        private void MoveActiveToValidPiece(int direction, int initial_move)
        {
            // Get the initial position
            int cur = (int)ActivePlayer.ActivePiece;
            cur = MathUtil.WrapRange(cur + initial_move, 0, (int)GamePieces.__LAST__);

            // Find until we're there
            while (!PlayerHasPiece(ActivePlayer.Index, (GamePieces)cur))
            {
                cur = MathUtil.WrapRange(cur + direction, 0, (int)GamePieces.__LAST__);
            }

            // Set it
            ActivePlayer.ActivePiece = (GamePieces)cur;

            // Get the wall/player piece we've changed to
            m_game_res.board.active_edit_piece = GetWallPiece(ActivePlayer.ActivePiece);
            m_game_res.board.active_player_piece = GetPlayerPiece(ActivePlayer.Index, ActivePlayer.ActivePiece);

            // Changed into a wall piece?
            if (m_game_res.board.active_edit_piece != null)
            {
                m_game_res.board.active_edit_piece.CurrentPlaceIsValid =
                    m_game_res.board.active_edit_piece.ValidOnBoard(m_game_res.board);
            }

            // Changed into a player piece?
            m_game_res.board.active_player_piece_destinations.Clear();
            if (m_game_res.board.active_player_piece != null)
            {
                m_game_res.board.GetPieceDestinations(
                    m_game_res.board.active_player_piece,
                    m_game_res.board.active_player_piece_destinations);
            }
        }

        private bool PlayerHasPiece(PlayerIndex plr, GamePieces piece)
        {
            if (piece < GamePieces.Button1)
            {
                return (m_players[(int)plr].PieceCount[(int)piece] > 0);
            }
            return (GetPlayerPiece(plr, piece) != null);
        }

        private WallPiece GetWallPiece(GamePieces piece)
        {
            if (piece >= GamePieces.Button1)
                return null;
            return WallPieces[(int)piece];
        }

        private PlayerPiece GetPlayerPiece(PlayerIndex plr, GamePieces piece)
        {
            if (piece < GamePieces.Button1)
                return null;

            PlayerPiece ret = m_game_res.board.GetPiece(plr, (int)piece - (int)GamePieces.Button1);
            if (ret == null || ret.Finished)
                return null;

            return ret;
        }

        private void ValidatePosition()
        {
            Point pos = ActivePlayer.PiecePosition;
            int max_x = Board.Width;
            int max_y = Board.Height;

            if (ActivePlayer.ActivePiece >= GamePieces.Button1)
            { // Different limits for player pieces
                max_x -= 1;
                max_y -= 1;
            }

            if (pos.X < 0) pos.X = 0;
            if (pos.Y < 0) pos.Y = 0;
            if (pos.X > max_x) pos.X = max_x;
            if (pos.Y > max_y) pos.Y = max_y;

            ActivePlayer.PiecePosition = pos;
        }

        private bool MakeMove()
        {
            Board board = m_game_res.board;

            if (board.active_edit_piece != null)
            { // Is the wall placement successful?
                if (board.active_edit_piece.PlaceOnBoard(board))
                {
                    // Make a sound
                    m_res.sound_bank.PlayCue("WallPlace");

                    // Deduct the piece from the player
                    ActivePlayer.PieceCount[(int)ActivePlayer.ActivePiece] -= 1;
                    return true;
                }
            }
            else if (board.active_player_piece != null)
            { // Can we move where we're moving to?
                if (board.active_player_piece_destinations.Contains(ActivePlayer.PiecePosition))
                {
                    // Make a sound
                    m_res.sound_bank.PlayCue("PieceMove");

                    // Make the move
                    board.active_player_piece.Position = ActivePlayer.PiecePosition;
                    return true;
                }
            }

            return false;
        }

        private void NextTurn()
        {
            Board board = m_game_res.board;

            // Reset the board pieces
            board.active_edit_piece = null;
            board.active_player_piece = null;
            board.active_player_piece_destinations.Clear();
            board.SaveUndo();

            // Player finished?
            if (board.HasPlayerCompleted(ActivePlayer.Index))
            {
                ActivePlayer.FinishPosition = m_finished_count;
                m_finished_count++;
            }

            // Game done?
            if (board.HasGameCompleted)
            {
                m_game_complete = true;
                return;
            }

            // Next unfinished player
            do
            {
                m_active_player = (m_active_player + 1) % m_players.Length;
            } while (board.HasPlayerCompleted((PlayerIndex)m_active_player) == true);

            // Init the turn
            InitTurn();
        }

        private Point? GetMouseBoardLocation(bool wall_mode)
        {
            // Get the mouse location
            Vector2 mouse_pos = m_res.controller.UnscaledMousePos;
            Vector3 near_pos = new Vector3(mouse_pos.X, mouse_pos.Y, 0.0f);
            Vector3 far_pos = new Vector3(mouse_pos.X, mouse_pos.Y, 1.0f);

            // Get the projision/view/world matrices
            Matrix mtx_proj = m_game_res.board_vis.ProjisionMatrix;
            Matrix mtx_camera = m_game_res.board_vis.CameraMatrix;
            Matrix mtx_world = Matrix.Identity;

            // Unproject the points
            Vector3 near_world = m_res.graphics.GraphicsDevice.Viewport.Unproject(near_pos, mtx_proj, mtx_camera, mtx_world);
            Vector3 far_world = m_res.graphics.GraphicsDevice.Viewport.Unproject(far_pos, mtx_proj, mtx_camera, mtx_world);

            // Create a ray
            Vector3 ray_dir = far_world - near_world;
            ray_dir.Normalize();
            Ray ray = new Ray(near_world, ray_dir);

            // Intersect with the plane
            Plane ground = new Plane(new Vector3(0, 1, 0), -0.25f);
            float? hit_dist;
            ray.Intersects(ref ground, out hit_dist);

            if (hit_dist == null)
                return null;

            // Get the hit position on the board
            Vector3 hit_pos = near_world + ray_dir * hit_dist.Value;
            hit_pos += new Vector3(Board.Width * 1.0f / 2.0f, 0.0f, Board.Height * 1.0f / 2.0f);
            if (!wall_mode)
                hit_pos += new Vector3(0.5f, 0.0f, 0.5f);
            else
                hit_pos += new Vector3(1.0f, 0.0f, 1.0f);

            Point board_pos = new Point((int)hit_pos.X, (int)hit_pos.Z);
            if (hit_pos.X < 0 ||
                hit_pos.Z < 0)
            {
                return null;
            }

            if (wall_mode && (board_pos.X > Board.Width || board_pos.Y > Board.Height))
                return null;
            if (!wall_mode && (board_pos.X >= Board.Width || board_pos.Y >= Board.Height))
                return null;

            return board_pos;
        }
    }
}
