using System;
using System.Collections.Generic;
using System.Text;
using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Graphics;
using Waboga;

namespace Waboga.UI
{
    public class MainUI
    {
        private Resources m_res;
        private GameResources m_game_res;

        private List<UIPage> m_pages = new List<UIPage>();

        private UIPage m_page_to_pop_until;
        private UIPage m_page_to_push;

        public UIPage TopPage
        {
            get
            {
                if (m_pages.Count <= 0)
                    return null;
                return m_pages[m_pages.Count - 1];
            }
        }

        public MainUI(Resources res, GameResources game_res)
        {
            m_res = res;
            m_game_res = game_res;

            // Start with the initial page
            //PushPage(new TopPage());
            PushPage(new MainMenuPage());
        }

        public void Deinit()
        {
            while (m_pages.Count > 0)
            {
                TopPage.Deinit();
                m_pages.RemoveAt(m_pages.Count - 1);
            }
        }

        public void Update(float delta_t)
        {
            // Pop if needed
            if (m_page_to_pop_until != null)
            {
                bool hide_page = true;
                int remove_count = 0;
                while (m_pages.Count > 0 && !object.ReferenceEquals(TopPage, m_page_to_pop_until))
                {
                    // Remove the page
                    if (hide_page)
                        TopPage.Hide();
                    TopPage.Deinit();

                    // Pop and continue
                    m_pages.RemoveAt(m_pages.Count - 1);

                    ++remove_count;
                    hide_page = false;
                }

                // Mark the top page as topmost
                if (remove_count <= 0)
                    throw new Exception("No pages popped!");
                if (m_pages.Count <= 0)
                    throw new Exception("Wanted page not found while poping the stack!");
                TopPage.IsTopmost = true;
                TopPage.Show();

                m_page_to_pop_until = null;
            }

            // Push a new page?
            if (m_page_to_push != null)
            {
                // Mark the old page as hidden
                if (TopPage != null)
                {
                    TopPage.IsTopmost = false;
                    TopPage.Hide();
                }

                // Init & add the new page
                m_pages.Add(m_page_to_push);
                TopPage.InitResources(m_res, m_game_res);
                TopPage.Init();
                TopPage.IsTopmost = true;
                TopPage.Show();

                m_page_to_push = null;
            }

            // Update from bottom->up
            for (int i = 0; i < m_pages.Count; ++i)
                m_pages[i].Update(this, delta_t);
        }

        public void Render()
        {
            // Create the scaling for 800x600 UI screen
            float ui_scale_x = (float)m_res.graphics.GraphicsDevice.Viewport.Width / 800.0f;
            float ui_scale_y = (float)m_res.graphics.GraphicsDevice.Viewport.Height / 600.0f;
            Matrix ui_scale_mtx = Matrix.CreateScale(ui_scale_x, ui_scale_y, 1.0f);

            m_res.controller.MousePosScale = new Vector2(1.0f / ui_scale_x, 1.0f / ui_scale_y);

            // Start the rendering
            m_res.ui_batch.Begin(
                SpriteBlendMode.AlphaBlend,
                SpriteSortMode.Immediate,
                SaveStateMode.SaveState,
                ui_scale_mtx);

            // Render from bottom->up
            for (int i = 0; i < m_pages.Count; ++i)
                m_pages[i].Render();

            // Render the mouse cursor
#if !XBOX
            if (!m_res.controller.MouseDragMode)
            {
                m_res.ui_batch.Draw(
                    m_res.tex_cursor,
                    m_res.controller.MousePos,
                    null,
                    Color.White,
                    0.0f,
                    new Vector2(5, 4),
                    0.3f,
                    SpriteEffects.None,
                    0.0f);
            }
#endif

            // End the rendering
            m_res.ui_batch.End();
        }

        public void PushPage(UIPage page)
        {
            if (page == null)
                throw new ArgumentNullException();
            m_page_to_push = page;
        }

        public void PopPage()
        {
            m_page_to_pop_until = m_pages[m_pages.Count - 2];
        }

        public void PopUntil(UIPage page)
        {
            m_page_to_pop_until = page;
        }

        public void PopUntil(Type type)
        {
            for (int i = m_pages.Count - 1; i >= 0; --i)
            {
                if (type.IsAssignableFrom(m_pages[i].GetType()))
                {
                    m_page_to_pop_until = m_pages[i];
                    return;
                }
            }

            throw new Exception("Would pop entire page stack!");
        }
    }
}
