using System;
using System.Collections.Generic;
using System.Text;
using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Graphics;
using Waboga;

namespace Waboga.UI
{
    public abstract class UIPage
    {
        protected Resources m_res;
        protected GameResources m_game_res;
        private UIRenderer m_renderer;

        protected Controller Controller
        {
            get { return m_res.controller; }
        }

        private bool m_is_topmost;
        public virtual bool IsTopmost
        {
            get { return m_is_topmost; }
            set { m_is_topmost = value; }
        }

        public virtual void InitResources(Resources res, GameResources game_res)
        {
            m_res = res;
            m_game_res = game_res;
            m_renderer = new UIRenderer(m_res);
        }

        public virtual void Init()
        {
        }

        public virtual void Deinit()
        {
        }

        public abstract void Update(MainUI ui, float delta_t);

        public virtual void Created()
        {
        }

        public virtual void Show()
        {
        }

        public virtual void Hide()
        {
        }

        public virtual void Render()
        {
            m_renderer.Render();
        }

        ////// UI STATE & UPDATE //////
        private String m_ui_active_item;
        private String m_ui_last_item;
        private bool m_ui_ctrl_move_up;
        private bool m_ui_ctrl_move_down;
        private bool m_ui_ctrl_move_left;
        private bool m_ui_ctrl_move_right;
        private bool m_ui_ctrl_select;

        /// <summary>
        /// Starts the round of UI update.
        /// </summary>
        public virtual UIRenderer UIBegin()
        {
            m_renderer.Begin();

            m_ui_ctrl_move_up =
                Controller.WasPressed(ControllerButton.PadUp) ||
                Controller.WasPressed(ControllerButton.LeftStickUp);
            m_ui_ctrl_move_down =
                Controller.WasPressed(ControllerButton.PadDown) ||
                Controller.WasPressed(ControllerButton.LeftStickDown);
            m_ui_ctrl_move_left =
                Controller.WasPressed(ControllerButton.PadLeft) ||
                Controller.WasPressed(ControllerButton.LeftStickLeft);
            m_ui_ctrl_move_right =
                Controller.WasPressed(ControllerButton.PadRight) ||
                Controller.WasPressed(ControllerButton.LeftStickRight);
            m_ui_ctrl_select =
                Controller.WasPressed(ControllerButton.A) ||
                Controller.WasPressed(ControllerButton.MouseLeft);

            if (!IsTopmost)
            {
                m_ui_ctrl_move_up = false;
                m_ui_ctrl_move_down = false;
                m_ui_ctrl_move_left = false;
                m_ui_ctrl_move_right = false;
                m_ui_ctrl_select = false;
            }

            return m_renderer;
        }

        public virtual bool UIButton(string id, string text)
        {
            // If no item is active, then activate
            if (m_ui_active_item == null)
                m_ui_active_item = id;

            // Item switching
            bool is_selected = false;
            if (m_ui_active_item == id)
            {
                if (m_ui_ctrl_move_down)
                {
                    m_ui_active_item = null;
                    m_res.sound_bank.PlayCue("MenuSelect");
                }
                else if (m_ui_ctrl_move_up)
                {
                    m_ui_active_item = m_ui_last_item;
                    m_res.sound_bank.PlayCue("MenuSelect");
                }
                else
                {
                    is_selected = m_ui_ctrl_select;
                    if (is_selected)
                        m_res.sound_bank.PlayCue("MenuSelect");
                }

                // Clear the keys
                UIClearKeys();
            }

            // Display the item
            Color col = (m_ui_active_item == id) ? Color.White : new Color(0x77, 0x77, 0x77);
            Rectangle rect = m_renderer.WriteLineCentered(text, col);

            // Mouse selecting
            if (MouseInRect(ref rect) && IsTopmost)
                m_ui_active_item = id;

            // Mark this as the last item
            m_ui_last_item = id;

            return is_selected;
        }

        public virtual void UIBoolean(string id, string text, ref bool value)
        {
            // If no item is active, then activate
            if (m_ui_active_item == null)
                m_ui_active_item = id;

            // Item switching
            bool change = false;
            if (m_ui_active_item == id)
            {
                if (m_ui_ctrl_move_down)
                {
                    m_ui_active_item = null;
                    m_res.sound_bank.PlayCue("MenuSelect");
                }
                else if (m_ui_ctrl_move_up)
                {
                    m_ui_active_item = m_ui_last_item;
                    m_res.sound_bank.PlayCue("MenuSelect");
                }
                else
                {
                    if (m_ui_ctrl_move_left ||
                        Controller.WasPressed(ControllerButton.MouseRight) ||
                        m_ui_ctrl_select ||
                        m_ui_ctrl_move_right)
                    {
                        change = true;
                    }
                }

                // Clear the keys
                UIClearKeys();
            }

            // Value changing
            if (change)
            {
                value = !value;
                m_res.sound_bank.PlayCue("MenuSelect");
            }

            // Display the item
            Color col = (m_ui_active_item == id) ? Color.White : new Color(0x77, 0x77, 0x77);
            string value_txt = value ? ": Yes" : ": No";
            Rectangle rect = m_renderer.WriteLineCentered(text + value_txt, col);

            // Mouse selecting
            if (MouseInRect(ref rect) && IsTopmost)
                m_ui_active_item = id;

            // Mark this as the last item
            m_ui_last_item = id;
        }

        public virtual void UINumber(string id, string text, int min, int max, ref int value)
        {
            // If no item is active, then activate
            if (m_ui_active_item == null)
                m_ui_active_item = id;

            // Item switching
            int change_amt = 0;
            if (m_ui_active_item == id)
            {
                if (m_ui_ctrl_move_down)
                {
                    m_ui_active_item = null;
                    m_res.sound_bank.PlayCue("MenuSelect");
                }
                else if (m_ui_ctrl_move_up)
                {
                    m_ui_active_item = m_ui_last_item;
                    m_res.sound_bank.PlayCue("MenuSelect");
                }
                else
                {
                    if (m_ui_ctrl_move_left || Controller.WasPressed(ControllerButton.MouseRight))
                        change_amt = -1;
                    if (m_ui_ctrl_select || m_ui_ctrl_move_right)
                        change_amt = 1;
                }

                // Clear the keys
                UIClearKeys();
            }

            // Value changing
            value = MathUtil.WrapRange(value + change_amt, min, max + 1);
            if (change_amt != 0)
                m_res.sound_bank.PlayCue("MenuSelect");

            // Display the item
            Color col = (m_ui_active_item == id) ? Color.White : new Color(0x77, 0x77, 0x77);
            Rectangle rect = m_renderer.WriteLineCentered(text + ": " + value, col);

            // Mouse selecting
            if (MouseInRect(ref rect) && IsTopmost)
                m_ui_active_item = id;

            // Mark this as the last item
            m_ui_last_item = id;
        }

        public virtual void UIIconNumber(string id, Texture2D texture, int min, int max, ref int value)
        {
            // If no item is active, then activate
            if (m_ui_active_item == null)
                m_ui_active_item = id;

            // Item switching
            int change_amt = 0;
            if (m_ui_active_item == id)
            {
                if (m_ui_ctrl_move_down)
                {
                    m_ui_active_item = null;
                    m_res.sound_bank.PlayCue("MenuSelect");
                }
                else if (m_ui_ctrl_move_up)
                {
                    m_ui_active_item = m_ui_last_item;
                    m_res.sound_bank.PlayCue("MenuSelect");
                }
                else
                {
                    if (m_ui_ctrl_move_left || Controller.WasPressed(ControllerButton.MouseRight))
                        change_amt = -1;
                    if (m_ui_ctrl_select || m_ui_ctrl_move_right)
                        change_amt = 1;
                }

                // Clear the keys
                UIClearKeys();
            }

            // Value changing
            value = MathUtil.WrapRange(value + change_amt, min, max + 1);
            if (change_amt != 0)
                m_res.sound_bank.PlayCue("MenuSelect");

            // Display the item
            Color col = (m_ui_active_item == id) ? Color.White : new Color(0x77, 0x77, 0x77);
            Rectangle rect = m_renderer.DrawIconAndTextCentered(texture, " " + value, col, 0.6f);

            // Mouse selecting
            if (MouseInRect(ref rect) && IsTopmost)
                m_ui_active_item = id;

            // Mark this as the last item
            m_ui_last_item = id;
        }

        public virtual bool UIBackButton()
        {
            bool back_pressed =
                Controller.WasPressed(ControllerButton.B) ||
                Controller.WasPressed(ControllerButton.KeyEsc);

            if (back_pressed && IsTopmost)
            {
                m_res.sound_bank.PlayCue("MenuSelect");
                return true;
            }
            return false;
        }

        /// <summary>
        /// Ends the round of UI update.
        /// </summary>
        public virtual void UIEnd()
        {
            m_renderer.End();
        }

        private void UIClearKeys()
        {
            m_ui_ctrl_move_up = false;
            m_ui_ctrl_move_down = false;
            m_ui_ctrl_move_left = false;
            m_ui_ctrl_move_right = false;
            m_ui_ctrl_select = false;
        }

        private bool MouseInRect(ref Rectangle rect)
        {
            Vector2 mpos = m_res.controller.MousePos;
            return rect.Contains((int)mpos.X, (int)mpos.Y);
        }
    }
}
