using System;
using System.Collections.Generic;
using System.Text;
using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Graphics;
using Waboga;

namespace Waboga.UI
{
    public class UIRenderer
    {
        private struct Window
        {
            public Vector2 area_min;
            public Vector2 area_max;
            public Vector2 write_pos;

            public float Width
            {
                get { return area_max.X - area_min.X; }
            }

            public float Height
            {
                get { return area_max.Y - area_min.Y; }
            }

            public void Reset(Vector2 start_pt, Vector2 end_pt)
            {
                area_min = start_pt;
                area_max = end_pt;
                write_pos = new Vector2(1.0f, 0.0f);
            }
        }
        private enum DrawEventType
        {
            None,
            FillRect,
            DrawText,
            DrawTexture,
        }
        private struct DrawEvent
        {
            public DrawEventType type;
            public Color color;

            public Rectangle fill_rect;

            public String draw_text;
            public Vector2 pos;
            public float scale;

            public Texture2D draw_texture;

            public void InitFillRect(Vector2 start_pt, Vector2 end_pt, Color col)
            {
                type = DrawEventType.FillRect;
                color = col;
                fill_rect = new Rectangle((int)start_pt.X, (int)start_pt.Y, (int)(end_pt.X - start_pt.X), (int)(end_pt.Y - start_pt.Y));
            }

            public void InitDrawText(String txt, Vector2 _pos, Color col)
            {
                type = DrawEventType.DrawText;
                color = col;
                draw_text = txt;
                pos = _pos;
            }

            public void InitDrawTexture(Texture2D texture, Vector2 _pos)
            {
                type = DrawEventType.DrawTexture;
                draw_texture = texture;
                pos = _pos;
            }
        }

        private const float FontScale = 0.6f;
        private static Color DefaultWindowColor = new Color(0, 0, 80, 50);

        private Resources m_res;

        private Window m_window;
        private Stack<Window> m_window_stack = new Stack<Window>();

        private List<DrawEvent> m_draw_events = new List<DrawEvent>();

        public UIRenderer(Resources res)
        {
            m_res = res;
        }

        public void Begin()
        {
            m_window.Reset(new Vector2(0.0f, 0.0f), new Vector2(800.0f, 600.0f));
            m_window_stack.Clear();

            m_draw_events.Clear();
        }

        public void End()
        {
        }

        public void WindowStart(Vector2 start_pt, Vector2 end_pt)
        {
            WindowStart(start_pt, end_pt, DefaultWindowColor);
        }

        public void WindowStart(Vector2 start_pt, Vector2 end_pt, Color fill_color)
        {
            m_window_stack.Push(m_window);

            // Create the new area
            Vector2 new_start = m_window.area_min + start_pt;
            Vector2 new_end = m_window.area_min + end_pt;
            m_window.Reset(new_start, new_end);

            // Add the fill event
            DrawEvent evt = new DrawEvent();
            evt.InitFillRect(new_start, new_end, fill_color);
            m_draw_events.Add(evt);
        }

        public void WindowStartCentered(float y, float width, float height)
        {
            WindowStartCentered(y, width, height, DefaultWindowColor);
        }

        public void WindowStartCentered(float y, float width, float height, Color fill_color)
        {
            float x = (m_window.Width - width) / 2.0f;

            WindowStart(new Vector2(x, y), new Vector2(x + width, y + height), fill_color);
        }

        public void WindowEnd()
        {
            m_window = m_window_stack.Pop();
        }

        public void WriteAt(Vector2 pos, string txt, Color col, float scale)
        {
            // Add the text event
            DrawEvent evt = new DrawEvent();
            evt.InitDrawText(txt, m_window.area_min + pos, col);
            evt.scale = scale;
            m_draw_events.Add(evt);
        }

        public void WriteLine(string txt, Color col)
        {
            WriteLine(txt, col, 1.0f);
        }

        public void WriteLine(string txt, Color col, float scale)
        {
            scale *= FontScale;
            Vector2 dims = m_res.font_main.MeasureString(txt) * scale;

            // Add the text event
            DrawEvent evt = new DrawEvent();
            evt.InitDrawText(txt, m_window.area_min + m_window.write_pos, col);
            evt.scale = scale;
            m_draw_events.Add(evt);

            // Advance
            m_window.write_pos.Y += Math.Max(m_res.font_main.LineSpacing * scale, dims.Y);
        }

        public Rectangle WriteLineCentered(string txt, Color col)
        {
            return WriteLineCentered(txt, col, 1.0f);
        }

        public Rectangle WriteLineCentered(string txt, Color col, float scale)
        {
            scale *= FontScale;
            Vector2 dims = m_res.font_main.MeasureString(txt) * scale;
            
            Vector2 center_add = new Vector2((m_window.Width - dims.X) / 2.0f, 0);
            Vector2 write_pos = m_window.area_min + m_window.write_pos + center_add;

            // Add the text event
            DrawEvent evt = new DrawEvent();
            evt.InitDrawText(txt, write_pos, col);
            evt.scale = scale;
            m_draw_events.Add(evt);

            // Advance
            m_window.write_pos.Y += Math.Max(m_res.font_main.LineSpacing * scale, dims.Y);

            // Return the enclosing rectangle
            return new Rectangle((int)write_pos.X, (int)write_pos.Y, (int)dims.X, (int)dims.Y);
        }

        public Rectangle DrawIconAndTextAt(Vector2 pos, Texture2D texture, string txt, Color col, float scale)
        {
            float orig_write_x = m_window.write_pos.X;

            float text_h = m_res.font_main.LineSpacing * scale;

            pos += m_window.area_min;

            int rect_x = (int)pos.X;
            int rect_y = (int)pos.Y;

            // Draw the icon
            float tex_w = (float)texture.Width * scale;
            float tex_h = (float)texture.Height * scale;

            DrawEvent evt = new DrawEvent();
            evt.InitDrawTexture(texture, pos + new Vector2(0, (text_h - tex_h)* 0.5f));
            evt.scale = scale;
            m_draw_events.Add(evt);

            // Write the text
            pos.X += tex_w;

            evt = new DrawEvent();
            evt.InitDrawText(txt, pos, col);
            evt.scale = scale;
            m_draw_events.Add(evt);

            // Return the rectangle
            Vector2 text_size = m_res.font_main.MeasureString(txt) * scale;
            return new Rectangle(rect_x, rect_y, (int)(tex_w + text_size.X), (int)text_h);
        }

        public Rectangle DrawIconAndTextCentered(Texture2D texture, string txt, Color col, float scale)
        {
            float tex_w = (float)texture.Width * scale;
            float pos_x = m_window.write_pos.X + (m_window.Width - tex_w) / 2.0f;
            float pos_y = m_window.write_pos.Y;

            Rectangle ret = DrawIconAndTextAt(new Vector2(pos_x, pos_y), texture, txt, col, scale);

            // Advance
            m_window.write_pos.Y += m_res.font_main.LineSpacing * scale;

            return ret;
        }

        public void DrawIconAtCentered(Texture2D texture, float y, float scale)
        {
            float tex_w = (float)texture.Width * scale;
            float pos_x = m_window.write_pos.X + (m_window.Width - tex_w) / 2.0f;

            Vector2 pos = new Vector2(pos_x, y);

            // Draw the icon
            DrawEvent evt = new DrawEvent();
            evt.InitDrawTexture(texture, m_window.area_min + pos);
            evt.scale = scale;
            m_draw_events.Add(evt);
        }

        public void Render()
        {
            SpriteBatch batch = m_res.ui_batch;

            Vector2 font_shadow_dist = new Vector2(2.0f, 2.0f);

            for (int i = 0; i < m_draw_events.Count; ++i)
            {
                DrawEvent evt = m_draw_events[i];

                switch (evt.type)
                {
                    default:
                    case DrawEventType.None:
                        break;

                    case DrawEventType.FillRect:
                        batch.Draw(m_res.tex_white, evt.fill_rect, evt.color);
                        break;

                    case DrawEventType.DrawText:
                        batch.DrawString(m_res.font_main, evt.draw_text, evt.pos - font_shadow_dist, new Color(0, 0, 0, evt.color.A), 0.0f, Vector2.Zero, evt.scale, SpriteEffects.None, 0.0f);
                        batch.DrawString(m_res.font_main, evt.draw_text, evt.pos + font_shadow_dist, new Color(0, 0, 0, evt.color.A), 0.0f, Vector2.Zero, evt.scale, SpriteEffects.None, 0.0f);
                        batch.DrawString(m_res.font_main, evt.draw_text, evt.pos, evt.color, 0.0f, Vector2.Zero, evt.scale, SpriteEffects.None, 0.0f);
                        break;

                    case DrawEventType.DrawTexture:
                        //batch.Draw(evt.draw_texture, evt.pos, null, Color.White, 0.0f, Vector2.Zero, evt.scale, SpriteEffects.None, 1.0f);
                        batch.Draw(evt.draw_texture, new Rectangle((int)evt.pos.X, (int)evt.pos.Y, (int)((float)evt.draw_texture.Width * evt.scale), (int)((float)evt.draw_texture.Height * evt.scale)), Color.White);
                        break;
                }
            }
        }
    }
}
