using System;
using System.Collections.Generic;
using System.Text;
using Microsoft.Xna.Framework;

namespace Waboga
{
    public class WallPiece
    {
        /// <summary>
        /// Relative starting points for walls going to the left.
        /// </summary>
        private Point[] m_horiz_walls;
        /// <summary>
        /// Relative starting points for walls going to down.
        /// </summary>
        private Point[] m_vert_walls;
        /// <summary>
        /// Relative points for the intersection pegs.
        /// </summary>
        private Point[] m_intersections;

        private int m_unique_angles;
        public int UniqueAngles
        {
            get { return m_unique_angles; }
        }


        public int RotatedHorizWallCount
        {
            get { return (m_angle & 1) == 0 ? m_horiz_walls.Length : m_vert_walls.Length; }
        }

        public int RotatedVertWallCount
        {
            get { return (m_angle & 1) == 0 ? m_vert_walls.Length : m_horiz_walls.Length; }
        }

        public int HorizWallCount
        {
            get { return m_horiz_walls.Length; }
        }

        public int VertWallCount
        {
            get { return m_vert_walls.Length; }
        }

        public int IntersectionCount
        {
            get { return m_intersections.Length; }
        }

        private int m_angle;
        public int Angle
        {
            get { return m_angle; }
            set
            {
                m_angle = value;
                while (m_angle < 0) m_angle += 4;
                while (m_angle >= 4) m_angle -= 4;
            }
        }

        /// <summary>
        /// Position for the central intersection point.
        /// </summary>
        public Point Position;

        public bool CurrentPlaceIsValid;

        private WallPiece(Point[] horiz_walls, Point[] vert_walls, int unique_angles)
        {
            m_horiz_walls = horiz_walls;
            m_vert_walls = vert_walls;
            m_unique_angles = unique_angles;

            // Create the intersections
            byte[,] isects = new byte[11, 11];
            foreach (Point pt in m_horiz_walls)
            {
                isects[5 + pt.X, 5 + pt.Y] += 1;
                isects[5 + pt.X + 1, 5 + pt.Y] += 1;
            }
            foreach (Point pt in m_vert_walls)
            {
                isects[5 + pt.X, 5 + pt.Y] += 1;
                isects[5 + pt.X, 5 + pt.Y + 1] += 1;
            }

            List<Point> isect_pts = new List<Point>();
            for (int y = 0; y < isects.GetLength(1); ++y)
            {
                for (int x = 0; x < isects.GetLength(0); ++x)
                {
                    if (isects[x, y] > 1)
                        isect_pts.Add(new Point(x - 5, y - 5));
                }
            }
            m_intersections = isect_pts.ToArray();
        }

        private WallPiece(Point[] horiz_walls, Point[] vert_walls, Point[] intersections)
        {
            m_horiz_walls = horiz_walls;
            m_vert_walls = vert_walls;
            m_intersections = intersections;
        }

        public void Reset()
        {
            m_angle = 0;
            Position.X = Board.Width / 2;
            Position.Y = Board.Height / 2;
        }

        public void ValidatePosition()
        {
            if (Position.X < 0)
                Position.X = 0;
            if (Position.Y < 0)
                Position.Y = 0;
            if (Position.X > Board.Width)
                Position.X = Board.Width;
            if (Position.Y > Board.Height)
                Position.Y = Board.Height;
        }

        public Point GetOrigHorizWallPos(int idx)
        {
            return m_horiz_walls[idx];
        }

        public Point GetOrigVertWallPos(int idx)
        {
            return m_vert_walls[idx];
        }

        public Point GetOrigIntersectPos(int idx)
        {
            return m_intersections[idx];
        }

        public Point GetRotatedHorizWallPos(int idx)
        {
            Point pt = (m_angle & 1) == 0 ? m_horiz_walls[idx] : m_vert_walls[idx];

            switch (m_angle)
            {
                default:
                case 0:
                    return new Point(Position.X + pt.X, Position.Y + pt.Y);

                case 1:
                    return new Point(Position.X - pt.Y - 1, Position.Y + pt.X);

                case 2:
                    return new Point(Position.X - pt.X - 1, Position.Y - pt.Y);

                case 3:
                    return new Point(Position.X + pt.Y, Position.Y - pt.X);
            }
        }

        public Point GetRotatedVertWallPos(int idx)
        {
            Point pt = (m_angle & 1) == 0 ? m_vert_walls[idx] : m_horiz_walls[idx];

            switch (m_angle)
            {
                default:
                case 0:
                    return new Point(Position.X + pt.X, Position.Y + pt.Y);

                case 1:
                    return new Point(Position.X - pt.Y, Position.Y + pt.X);

                case 2:
                    return new Point(Position.X - pt.X, Position.Y - pt.Y - 1);

                case 3:
                    return new Point(Position.X + pt.Y, Position.Y - pt.X - 1);
            }
        }

        public Point GetRotatedIntersectPos(int idx)
        {
            Point pt = m_intersections[idx];

            switch (m_angle)
            {
                default:
                case 0:
                    return new Point(Position.X + pt.X, Position.Y + pt.Y);

                case 1:
                    return new Point(Position.X - pt.Y, Position.Y + pt.X);

                case 2:
                    return new Point(Position.X - pt.X, Position.Y - pt.Y);

                case 3:
                    return new Point(Position.X + pt.Y, Position.Y - pt.X);
            }
        }

        public bool ValidOnBoard(Board b)
        {
            // Test that all walls/intersections are inside the board and unoccupied
            for (int i = 0; i < RotatedHorizWallCount; ++i)
            {
                Point pt = GetRotatedHorizWallPos(i);
                if (pt.X < 0 || pt.X >= Board.Width)
                    return false;
                if (pt.Y < 0 || pt.Y > Board.Height)
                    return false;

                if (b.wall_horiz[pt.X, pt.Y])
                    return false;
            }
            for (int i = 0; i < RotatedVertWallCount; ++i)
            {
                Point pt = GetRotatedVertWallPos(i);
                if (pt.X < 0 || pt.X > Board.Width)
                    return false;
                if (pt.Y < 0 || pt.Y >= Board.Height)
                    return false;

                if (b.wall_vert[pt.X, pt.Y])
                    return false;
            }
            for (int i = 0; i < IntersectionCount; ++i)
            {
                Point pt = GetRotatedIntersectPos(i);
                if (pt.X < 0 || pt.X > Board.Width)
                    return false;
                if (pt.Y < 0 || pt.Y > Board.Height)
                    return false;

                if (b.filled_intersection[pt.X, pt.Y])
                    return false;
            }

            // Test that the board can be completed
            b.SaveUndo();
            UncheckedPlaceOnBoard(b);
            bool can_be_completed = b.CanBeCompleted;
            b.DoUndo();

            if (!can_be_completed)
                return false;

            return true;
        }

        public bool PlaceOnBoard(Board b)
        {
            if (!ValidOnBoard(b))
                return false;
            UncheckedPlaceOnBoard(b);
            return true;
        }

        private void UncheckedPlaceOnBoard(Board b)
        {
            for (int i = 0; i < RotatedHorizWallCount; ++i)
            {
                Point pt = GetRotatedHorizWallPos(i);
                b.wall_horiz[pt.X, pt.Y] = true;
            }
            for (int i = 0; i < RotatedVertWallCount; ++i)
            {
                Point pt = GetRotatedVertWallPos(i);
                b.wall_vert[pt.X, pt.Y] = true;
            }
            for (int i = 0; i < IntersectionCount; ++i)
            {
                Point pt = GetRotatedIntersectPos(i);
                b.filled_intersection[pt.X, pt.Y] = true;
            }
        }

        // 1 piece straight
        public static WallPiece PieceStraight1 = new WallPiece(
            new Point[] { new Point(-1, 0) },
            new Point[] { },
            2);

        // 2 piece straight
        public static WallPiece PieceStraight2 = new WallPiece(
            new Point[] { new Point(-1, 0), new Point(0, 0) },
            new Point[] { },
            2);

        // 3 piece straight
        public static WallPiece PieceStraight3 = new WallPiece(
            new Point[] { new Point(-1, 0), new Point(0, 0), new Point(1, 0) },
            new Point[] { },
            3);

        // 2 piece corner
        public static WallPiece PieceCorner = new WallPiece(
            new Point[] { new Point(-1, 0) },
            new Point[] { new Point(0, 0) },
            4);

        // 4 piece cross
        public static WallPiece PieceCross = new WallPiece(
            new Point[] { new Point(-1, 0), new Point(0, 0) },
            new Point[] { new Point(0, -1), new Point(0, 0) },
            1);

        // The tetris S-shape
        public static WallPiece PieceS1 = new WallPiece(
            new Point[] { new Point(-1, 0), new Point(0, 1) },
            new Point[] { new Point(0, 0) },
            2);

        // The tetris S-shape, reversed
        public static WallPiece PieceS2 = new WallPiece(
            new Point[] { new Point(-1, 1), new Point(0, 0) },
            new Point[] { new Point(0, 0) },
            2);

        // The L-shape
        public static WallPiece PieceL1 = new WallPiece(
            new Point[] { new Point(-1, 1), new Point(0, 1) },
            new Point[] { new Point(1, 0) },
            4);

        // The L-shape, reversed
        public static WallPiece PieceL2 = new WallPiece(
            new Point[] { new Point(-1, 1), new Point(0, 1) },
            new Point[] { new Point(-1, 0) },
            4);
    }
}
